VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "clsDpiImaging"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

' If false, removes 7 public methods & 9 private methods are removed or shortcircuited
#Const AllowThunkCapability = True      ' set to false to have class compile without
                                        ' any thunk/management support
#Const ThunkReleaseTesting = False      ' must be false unless testing thunk cleanup
                                        ' ^^ see pvCreateThunk for usage

' Purpose: Provide GDI+ scaling and rendering for StdPicture objects.
' Optional thunk can be used as needed. Thunk not absolutely required

'--------------------------------------------------------------------------------
'   WARNING     WARNING     WARNING     WARNING     WARNING     WARNING
'--------------------------------------------------------------------------------
'   The only known scenario where used thunk is not IDE safe is when a picture
'       is shared with another control and the picture to be shared is assigned
'       with control's default property, not the property name. So, don't do that
'   For example, if sharing Image1.Picture with Image2, you should always ensure
'       non-default properties are used, i.e. Image2.Picture = Image1.Picture
'   DO NOT share like this: Image2 = Image1 nor: Image2.Picture = Image1
'   Always use the property name. Thunk relies on calls to picture's IUnknown
'   method to track reference counting. But in above case, VB seems to bypass
'   that method which makes the thunk believe the ref count is greater than it
'   really is & fails to release properly.
'--------------------------------------------------------------------------------
'   GDI+ is a requirement for this class & picture manipulation. A thunk is not.
'   Thunk usage would NOT be the norm, instead is a tool for special cases
'   To thunk or not to thunk ... question answered
'       thunk: Transparency to be drawn over a non-solid background
'           Non-thunk: Provide a fill color for the picture's transparency
'       thunk: Dynamic GDI+ attributes to be applied, i.e., fading image on timer
'           Non-thunk: recreate image with attributes drawn into picture & filled
'       thunk: Dynamic high quality scaling
'           Non-thunk: recreate image each time size changes & fill as needed
'       thunk: animate GIFs
'           Non-thunk: no reasonable substitute
'       thunk: draw known alpha-capable bitmap to any hDC with transparency
'           Non-thunk: use AlphaBlend API (but alpha must be pARGB)
'           force pARGB, if needed, using these steps if thunk not used
'           1. hBitmap = thisClass.CopyPictureToHandle(thePicture, True)
'           2. Use AlphaBlend API with hBitmap
'           3. Use DeleteObject API to destroy hBitmap
'--------------------------------------------------------------------------------
'   Thunk is included for rendering transparency in icon, 32bppPARGB bitmaps,
'       PNG, TIF and optionally GIFs. Additionally, thunk is used for high
'       quality real-time scaling of any bitmap. The thunk is self-cleansing
'       and only one thunk needs to exists at any one time, for entire project.
'   The key point to remember about the thunk is that it is only effective for
'       managed pictures when that picture's Render method is called. That
'       method is called when picture exists in Image & PictureBox conrols and
'       the Form's picture property. It is not called during VB's PaintPicture,
'       rendering from ImageList controls, nor any other VB picture property.
'   This class does not need to be kept alive for a thunk to survive.
'   When thunk is used, a GDI+ image object is tied to the managed picture. This
'   adds little overhead because the GDI+ image uses the same pixel data/memory
'   that the picture uses. GIFs with enabled frame-navigation are the exception.
'--------------------------------------------------------------------------------
' Methods overview. Managed Pictures (MPs) means picture is thunk-rendered
' -- CopyPicture,LoadPictureEx,LoadResPictureEx are the key methods
'(*) methods do not exist if class is compiled without thunk/managment support
' * Attach: create a thunk and return its object, enables picture management
'   CopyPicture: scales picture while optionally applying other rendering effects
'   CopyPictureToHandle: create HICON or HBITMAP from existing picture
'   CreateGDIpAttributesHandle: courtesy method for common rendering options
'   DeleteGDIpAttributesHandle: self-explanatory
' * GdipImageAttributes: set/get attributes applied during rendering for MPs only
' * GdipScalingQuality: set/get rendering scaling quality for MPs only
'   GetAlphaFormat: returns pixel format of alpha-capable images
' * GetGDIpHandle: returns GDI+ handle used for MPs
'   GetImageDimensions: substitute for ScaleX/Y(Dimension, vbHimetric, vbPixels)
' * IsPictureManaged: test if picture is MP
'   LoadPictureEx: similar to VB's LoadPicture
'   LoadResPictureEx: similar to VB's LoadResPicture & LoadResData
' * ManagePicture: manage a previously unmanaged picture, requires thunk
'   RenderPicture: draws Picture, MP or not, to any hDC
' * UnmanagePicture: manually/force unmanagement of an existing MP
'   WrapHandleToStdPicture: return a stdPicture having same handle as passed
' ------------------------------------------------------------------------------

' ------------------------------------------------------------------------------
#If AllowThunkCapability = True Then ' thunk-only related event, enums, & APIs
Public Event Attach()
' This event is raised to remind you that a thunk needs to be assigned to the class.
' The event is only raised if picture management is about to happen, but thunk not assigned
' To get this optional event, declare this class instance using "WithEvents"
' See the class Attach method for more info

Private Declare Function GetModuleHandleA Lib "kernel32.dll" (ByVal lpModuleName As String) As Long
Private Declare Function GetProcAddress Lib "kernel32.dll" (ByVal hModule As Long, ByVal lpProcName As String) As Long
Private Declare Function VirtualAlloc Lib "kernel32.dll" (ByVal lpAddress As Long, ByVal dwSize As Long, ByVal flAllocationType As Long, ByVal flProtect As Long) As Long
Private Declare Function CallThunkProc Lib "user32.dll" Alias "CallWindowProcA" (ByVal lpPrevWndFunc As Long, ByVal hWnd As Long, ByVal msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
Private Declare Function CoTaskMemAlloc Lib "ole32.dll" (ByVal cb As Long) As Long
Private Declare Sub CoTaskMemFree Lib "ole32.dll" (ByVal pv As Long)
Private Declare Function DispCallFunc Lib "oleaut32.dll" (ByVal pvInstance As Long, ByVal offsetinVft As Long, ByVal CallConv As Long, ByVal retTYP As Integer, ByVal paCNT As Long, ByVal paTypes As Long, ByVal paValues As Long, ByRef retVAR As Variant) As Long

' management-only related enum
Public Enum ScalingQuality ' applies to managed pictures only
    sqNearestNeighbor = 0  ' see GdipScalingQuality property
    sqBiLinear
    sqBiCubic              ' default for newly managed pictures
End Enum
#End If
' ------------------------------------------------------------------------------

Private Declare Function CloseHandle Lib "kernel32.dll" (ByVal hObject As Long) As Long
Private Declare Function CreateFileW Lib "kernel32.dll" (ByVal lpFileName As Long, ByVal dwDesiredAccess As Long, ByVal dwShareMode As Long, ByRef lpSecurityAttributes As Long, ByVal dwCreationDisposition As Long, ByVal dwFlagsAndAttributes As Long, ByVal hTemplateFile As Long) As Long
Private Declare Sub CopyMemory Lib "kernel32.dll" Alias "RtlMoveMemory" (ByRef Destination As Any, ByRef Source As Any, ByVal Length As Long)
Private Declare Function GetFileSize Lib "kernel32.dll" (ByVal hFile As Long, ByRef lpFileSizeHigh As Long) As Long
Private Declare Function GlobalAlloc Lib "kernel32.dll" (ByVal uFlags As Long, ByVal dwBytes As Long) As Long
Private Declare Function GlobalFree Lib "kernel32.dll" (ByVal hMem As Long) As Long
Private Declare Function GlobalLock Lib "kernel32.dll" (ByVal hMem As Long) As Long
Private Declare Function GlobalSize Lib "kernel32.dll" (ByVal hMem As Long) As Long
Private Declare Function GlobalUnlock Lib "kernel32.dll" (ByVal hMem As Long) As Long
Private Declare Function ReadFile Lib "kernel32.dll" (ByVal hFile As Long, ByRef lpBuffer As Any, ByVal nNumberOfBytesToRead As Long, ByRef lpNumberOfBytesRead As Long, ByRef lpOverlapped As Any) As Long

Private Declare Function CopyImage Lib "user32.dll" (ByVal hHandle As Long, ByVal un1 As Long, ByVal n1 As Long, ByVal n2 As Long, ByVal un2 As Long) As Long
Private Declare Function CreateIconFromResourceEx Lib "user32.dll" (ByRef presbits As Any, ByVal dwResSize As Long, ByVal fIcon As Long, ByVal dwVer As Long, ByVal cxDesired As Long, ByVal cyDesired As Long, ByVal Flags As Long) As Long
Private Declare Function CreateIconIndirect Lib "user32.dll" (ByRef piconinfo As Any) As Long
Private Declare Function GetDC Lib "user32.dll" (ByVal hWnd As Long) As Long
Private Declare Function GetIconInfo Lib "user32.dll" (ByVal hIcon As Long, ByRef piconinfo As Any) As Long
Private Declare Function GetSysColor Lib "user32.dll" (ByVal nIndex As Long) As Long
Private Declare Function GetSystemMetrics Lib "user32.dll" (ByVal nIndex As Long) As Long
Private Declare Function ReleaseDC Lib "user32.dll" (ByVal hWnd As Long, ByVal hDC As Long) As Long

Private Declare Function CreateStreamOnHGlobal Lib "ole32.dll" (ByVal hGlobal As Long, ByVal fDeleteOnRelease As Long, ppstm As Any) As Long
Private Declare Function OleLoadPicture Lib "oleaut32.dll" (pStream As Any, ByVal lSize As Long, ByVal fRunmode As Long, riid As Any, ppvObj As Any) As Long
Private Declare Function OleCreatePictureIndirect Lib "oleaut32.dll" (lpPictDesc As Any, riid As Any, ByVal fPictureOwnsHandle As Long, iPic As IPicture) As Long
Private Declare Function IStream_Reset Lib "shlwapi.dll" (ByVal IStream As stdole.IUnknown) As Long
Private Declare Function VarPtrArray Lib "msvbvm60.dll" Alias "VarPtr" (var() As Any) As Long

Private Declare Function CreateBitmap Lib "gdi32.dll" (ByVal nWidth As Long, ByVal nHeight As Long, ByVal nPlanes As Long, ByVal nBitCount As Long, ByRef lpBits As Any) As Long
Private Declare Function CreateCompatibleDC Lib "gdi32.dll" (ByVal hDC As Long) As Long
Private Declare Function CreateDIBSection Lib "gdi32.dll" (ByVal hDC As Long, ByRef pBitmapInfo As Any, ByVal un As Long, ByRef lplpVoid As Long, ByVal Handle As Long, ByVal dw As Long) As Long
Private Declare Function DeleteDC Lib "gdi32.dll" (ByVal hDC As Long) As Long
Private Declare Function DeleteObject Lib "gdi32.dll" (ByVal hObject As Long) As Long
Private Declare Function GetDeviceCaps Lib "gdi32.dll" (ByVal hDC As Long, ByVal nIndex As Long) As Long
Private Declare Function GetDIBits Lib "gdi32.dll" (ByVal aHDC As Long, ByVal hBitmap As Long, ByVal nStartScan As Long, ByVal nNumScans As Long, ByRef lpBits As Any, ByRef lpBI As Any, ByVal wUsage As Long) As Long
Private Declare Function GetGDIObject Lib "gdi32.dll" Alias "GetObjectA" (ByVal hObject As Long, ByVal nCount As Long, ByRef lpObject As Any) As Long
Private Declare Function SelectObject Lib "gdi32.dll" (ByVal hDC As Long, ByVal hObject As Long) As Long

Private Declare Function GdipBitmapLockBits Lib "GdiPlus.dll" (ByVal mBitmap As Long, mRect As Any, ByVal mFlags As Long, ByVal mPixelFormat As Long, mLockedBitmapData As Any) As Long
Private Declare Function GdipBitmapUnlockBits Lib "GdiPlus.dll" (ByVal mBitmap As Long, mLockedBitmapData As Any) As Long
Private Declare Function GdipCloneImageAttributes Lib "GdiPlus.dll" (ByVal imageattr As Long, ByRef imageattrClone As Long) As Long
Private Declare Function GdipCreateBitmapFromHBITMAP Lib "GdiPlus.dll" (ByVal pbitmap As Long, ByVal hPal As Long, ByRef hbmReturn As Long) As Long
Private Declare Function GdipCreateBitmapFromScan0 Lib "GdiPlus.dll" (ByVal Width As Long, ByVal Height As Long, ByVal stride As Long, ByVal PixelFormat As Long, scan0 As Any, BITMAP As Long) As Long
Private Declare Function GdipCreateFromHDC Lib "GdiPlus.dll" (ByVal hDC As Long, hGraphics As Long) As Long
Private Declare Function GdipCreateHBITMAPFromBitmap Lib "GdiPlus.dll" (ByVal pbitmap As Long, ByRef hBmp As Long, ByVal Color As Long) As Long
Private Declare Function GdipCreateHICONFromBitmap Lib "GdiPlus.dll" (ByVal pbitmap As Long, ByRef hIcon As Long) As Long
Private Declare Function GdipCreateImageAttributes Lib "GdiPlus.dll" (ByRef imageattr As Long) As Long
Private Declare Function GdipDeleteGraphics Lib "GdiPlus.dll" (ByVal mGraphics As Long) As Long
Private Declare Function GdipDisposeImage Lib "GdiPlus.dll" (ByVal Image As Long) As Long
Private Declare Function GdipDisposeImageAttributes Lib "GdiPlus.dll" (ByVal imageattr As Long) As Long
Private Declare Function GdipDrawImageRectRectI Lib "GdiPlus.dll" (ByVal hGraphics As Long, ByVal hImage As Long, ByVal DstX As Long, ByVal DstY As Long, ByVal DstWidth As Long, ByVal DstHeight As Long, ByVal SrcX As Long, ByVal SrcY As Long, ByVal SrcWidth As Long, ByVal SrcHeight As Long, ByVal srcUnit As Long, ByVal imageAttributes As Long, ByVal Callback As Long, ByVal callbackData As Long) As Long
Private Declare Function GdipGetImageBounds Lib "GdiPlus.dll" (ByVal nImage As Long, srcRect As Any, srcUnit As Long) As Long
Private Declare Function GdipGetImageGraphicsContext Lib "GdiPlus.dll" (ByVal pImage As Long, ByRef graphics As Long) As Long
Private Declare Function GdipGetImageHorizontalResolution Lib "GdiPlus.dll" (ByVal pImage As Long, ByRef resolution As Single) As Long
Private Declare Function GdipGetImagePixelFormat Lib "GdiPlus.dll" (ByVal hImage As Long, PixelFormat As Long) As Long
Private Declare Function GdipGetImageType Lib "GdiPlus.dll" (ByVal hImage As Long, ByRef nType As Long) As Long
Private Declare Function GdipGetImageVerticalResolution Lib "GdiPlus.dll" (ByVal pImage As Long, ByRef resolution As Single) As Long
Private Declare Function GdipGetPropertyItem Lib "GdiPlus.dll" (ByVal pImage As Long, ByVal propId As Long, ByVal propSize As Long, ByRef buffer As Any) As Long
Private Declare Function GdipGetPropertyItemSize Lib "GdiPlus.dll" (ByVal pImage As Long, ByVal propId As Long, ByRef pSize As Long) As Long
Private Declare Function GdipGraphicsClear Lib "GdiPlus.dll" (ByVal graphics As Long, ByVal pColor As Long) As Long
Private Declare Function GdipImageGetFrameCount Lib "GdiPlus.dll" (ByVal Image As Long, ByRef dimensionID As Any, ByRef Count As Long) As Long
Private Declare Function GdipImageRotateFlip Lib "GdiPlus.dll" (ByVal pImage As Long, ByVal rfType As Long) As Long
Private Declare Function GdipImageSelectActiveFrame Lib "GdiPlus.dll" (ByVal Image As Long, ByRef dimensionID As Any, ByVal FrameIndex As Long) As Long
Private Declare Function GdipLoadImageFromStream Lib "GdiPlus.dll" (ByVal Stream As Long, Image As Long) As Long
Private Declare Function GdiplusShutdown Lib "GdiPlus.dll" (Token As Long) As Long
Private Declare Function GdiplusStartup Lib "GdiPlus.dll" (Token As Long, inputbuf As Any, Optional ByVal outputbuf As Long = 0) As Long
Private Declare Function GdipSetImageAttributesColorMatrix Lib "GdiPlus.dll" (ByVal imgAttr As Long, ByVal clrAdjust As Long, ByVal clrAdjustEnabled As Long, ByRef clrMatrix As Any, ByRef grayMatrix As Any, ByVal clrMatrixFlags As Long) As Long
Private Declare Function GdipSetImageAttributesWrapMode Lib "GdiPlus.dll" (ByVal imageattr As Long, ByVal wrap As Long, ByVal pColor As Long, ByVal clamp As Long) As Long
Private Declare Function GdipSetInterpolationMode Lib "GdiPlus.dll" (ByVal hGraphics As Long, ByVal Interpolation As Long) As Long
Private Declare Function GdipSetPixelOffsetMode Lib "GdiPlus.dll" (ByVal hGraphics As Long, ByVal Interpolation As Long) As Long

Public Enum AlphaUsageEnum
    ' uses 0x7F000000 for enum; they won't conflict with RGB/system color values
    acApplyAlphaChannel = 0
    acIgnoreAlphaChannel = &H1000000 ' can also be OR'd with color key
    acWantFormat32bppPARGB = &H2000000 ' can also be OR'd with color key
    acColorKeyIsTopLeftPixel = &H4000000
    acColorKeyIsBlack = &H8000000   ' since 0 (black) used for acApplyAlphaChannel
    ' else use any valid RGB/system color value to act as colorkey
    ' coder notes:
    '   enum values are masked with: 0x8FFFFFFF, unused bits (0x70000000) are masked out
End Enum
Public Enum ManagedStyle ' used in several methods
    msAutoSelect = 0    ' selects logical choice based on method's parameters
    msUnmanagedBitmap   ' conversion to standard bitmap
    msManagedBitmap     ' conversion to transparency-capable, thunk-required bitmap
End Enum
Public Enum AlphaTypeEnum   ' used for GetAlphaFormat only
    alphaNone = 0       ' alpha channel not used or non-existant
    alphaARGB           ' alpha channel is ARGB (similar to icons)
    alphaPARGB          ' alpha channel is PARGB (compatible with AlphaBlend API)
    alphaUniversal      ' alpha is both ARGB & PARGB (similar to GIFs)
End Enum
Public Enum ColorScaleRatioEnum ' used for CreateGDIpAttributesHandle only
    gsNone = 0          ' no grayscaling
    gsNTSCPAL           ' R*.299, G*.587, B*.114
    gsCCIR709           ' R*.213, G*.715, B*.072
    gsSimpleAverage     ' RGB = (R+G+B)/3
    gsRedMask           ' uses only the Red sample value: G,B = 0
    gsGreenMask         ' uses only the Green sample value: R,B = 0
    gsBlueMask          ' uses only the Blue sample value: R,G = 0
    gsRedGreenMask      ' uses Red & Green sample value: R*.5,G*.5,B=0
    gsBlueGreenMask     ' uses Blue & Green sample value: R=0,G*.5,B*.5
    gsAlphaMask         ' R=alpha,G=alpha,B=alpha
    gsSepia             ' mixture of RGB to create a brown-scale effect
End Enum

Private Enum eAlphaType ' internal use
    PixelFormat24bppRGB = &H21808   ' non-alpha, minimal depth for managed pictures
    PixelFormat32bppRGB = &H22009   ' non-alpha
    PixelFormat32bppPARGB = &HE200B
    PixelFormat32bppARGB = &H26200A
    PixelFormat32bppUARGB = PixelFormat32bppARGB Or PixelFormat32bppPARGB
    alphaExists = PixelFormat32bppPARGB ' formats less than this: not alpha format
End Enum
Private Enum ePicType   ' internal use
    typeUnknown = 0
    typeBitmap
    typeJPG
    typeGIF
    typeIcon
    typeCursor
    typeMetafile
    typePNG
    typeTIFF
End Enum
Private Type BitmapData
    ' 2 combined GDI+ structures used for reading/writing GDI+ pixel data
    bpLeft As Long          ' {GDI+ RECTI begins here...
    bpTop As Long
    bpWidth As Long         ' {GDI+ BitmapData begins here...
    bpHeight As Long        ' } GDI+ RECTI final member
    bpStride As Long
    bpPixelFormat As Long
    bpScan0Ptr As Long
    bpReserved As Long      ' } GDI+ BitmapData final member
End Type

Const UnitPixels As Long = 2            ' GDI+ scalemode
Const PICTURE_TRANSPARENT = 2           ' VB IPicture.Atributes flag
Const WrapModeTileFlipXY As Long = 3    ' GDI+ Image Attributes option
Const AlphaUsageMask As Long = &H8FFFFFFF ' mask to exclude invalid bits
Const Offset_ClientPropA As Long = 8    ' offset from managed picture's data array
Const Offset_ClientPropB As Long = 12   ' offset from managed picture's data array
Const Offset_ClientList As Long = 8     ' offset from canary ObjPtr
Const Offset_Unhook As Long = &H78      ' offset to thunk's Unhook method

Dim m_GdipToken As Long
Dim m_Canary As stdole.IUnknown         ' object related to an active thunk
Dim m_DPI As Long                       ' system DPI

#If AllowThunkCapability = True Then    ' management-only
'/////// following public methods are compiled only for thunk/management support

Public Function Attach(oStdPictureEx As stdole.IUnknown) As Boolean

    ' If never called, pass an uninstantiated IUnknown else pass
    '   the IUnknown returned from a previous call to this method.
    ' An active thunk is required for these methods when management applies:
    '   CopyPicture, LoadPictureEx, LoadResPictureEx, ManagePicture
    
    ' The returned parameter must be cached to enable picture management.
    ' You should cache the returned object in your main form or in a module
    '   that will stay around as long as you want management. This class
    '   does not need to be kept alive. The class can be created as needed.
    '   The thunk created by the class stays alive as long as the passed
    '   parameter does.
    '   --------------------------------------------------------------
    '   Dim oStdPictureEx As stdole.IUnknown ' declarations section
    '   -- as needed elsewhere in your code:
    '       Dim cImaging As clsDpiImaging
    '       Set cImaging = New clsDpiImaging
    '       cImaging.Attach oStdPictureEx
    
    '   Alternate method, using the optional Attach event & WithEvents
    '   --------------------------------------------------------------
    '   Dim oStdPictureEx As stdole.IUnknown ' declarations section
    '   Dim WithEvents cImaging As clsDpipImaging ' also declarations section
    '   Private Sub cImaging_Attach() ' event called as needed
    '       cImaging.Attach oStdPictureEx
    '   End Sub
    '   -- as needed elsewhere in your code:
    '       If cImaging Is Nothing Then Set cImaging = New clsDpiImaging
    '       ... call methods of that class
    
    '   --------------------------------------------------------------
    ' If the passed parameter goes out of scope or is set to Nothing, then
    ' all existing managed pictures related to that parameter are unmanaged.
    ' This class keeps a reference, but if the class is unloaded and later
    '   reloaded, ensure you call this method afterwards.
    
    ' Not recommended, but you could create multiple oStdPictureEx objects.
    ' Each one will result in a newly created thunk. If you use multiple objects,
    ' ensure this class has the one you want to use before you manage any new
    ' pictures. Newly managed pictures always use the reference held by the class.

    Dim oCanary As stdole.IUnknown
    If oStdPictureEx Is Nothing Then    ' passed item is not instantiated
        If pvCreateCanary(oCanary) = True Then
            Set oStdPictureEx = oCanary: Attach = True
        End If
    Else                                ' update class to passed item
        Set oCanary = pvGetCanary(oStdPictureEx)
        Attach = (Not oCanary Is Nothing)
    End If
    Set m_Canary = oCanary: Set oCanary = Nothing

End Function

Public Function ManagePicture(thePicture As IPicture, _
                    Optional ByVal FillColor As Long = -1, _
                    Optional ByVal BmpAlphaUsage As AlphaUsageEnum = acApplyAlphaChannel) As StdPicture
                        
    ' Set Image1.Picture = thisClass.ManagePicture(Image1.Picture)
    ' Image1.Container.Refresh ' may be needed if returned picture is not the passed picture
    
    ' Manages an unmanaged picture. Managing always converts image formats to bitmap.
    
    '   FillColor: optional opaque color to replace transparency, VB system colors accepted
    '       -1 (default) = allow transparency, will be black for unmanaged pictures.
    '       Otherwise, pass valid color to replace the transparency within the image.
    '   AlphaUsage:
    '      acApplyAlphaChannel (default) = test for alpha channel & apply if it exists
    '      acIgnoreAlphaChannel = alpha not used. Set this for VB Clipboard pictures
    '           acIgnoreAlphaChannel can be combined with color keys
    '      acWantFormat32bppPARGB = return PARGB format if transparency exists
    '           acWantFormat32bppPARGB can be combined with color keys
    '      acColorKeyIsTopLeftPixel = all pixels same as top/left pixel color are transparent
    '      acColorKeyIsBlack = color black is transparent
    '      else any valid color = make transparent all pixels equal to this color
    '(*) note that if no thunk is created, any managed pictures that have transparency,
    '    will have transparency rendered as black without FillColor. Call the Attach method.
    
    ' The returned picture will not be the same picture passed to this routine
                        
    If pvIsPictureValid(thePicture) = False Then Exit Function
    If Not pvGetCanary(thePicture) Is Nothing Then GoTo exitRoutine ' already managed
    If pvCanaryExists() = False Then GoTo exitRoutine
    
    Dim tPic As StdPicture
    
    If FillColor <> -1 Then
         If FillColor < 0 Then FillColor = GetSysColor(FillColor And &HFF)
    End If
    BmpAlphaUsage = BmpAlphaUsage And AlphaUsageMask
    If (BmpAlphaUsage And (acIgnoreAlphaChannel Or acWantFormat32bppPARGB)) = (acIgnoreAlphaChannel Or acWantFormat32bppPARGB) Then
        BmpAlphaUsage = BmpAlphaUsage Xor acWantFormat32bppPARGB
    End If
    
    Select Case thePicture.Type
    Case vbPicTypeEMetafile, vbPicTypeMetafile
        Set tPic = pvConvertMetafile(thePicture, 0, 0, FillColor, msManagedBitmap, BmpAlphaUsage, 0)
    Case vbPicTypeBitmap
        If (thePicture.Attributes And PICTURE_TRANSPARENT) = PICTURE_TRANSPARENT Then ' GIF
            Set tPic = pvConvertGIF(thePicture, 0, 0, FillColor, msManagedBitmap, BmpAlphaUsage, 0)
        Else
            Set tPic = pvProcessBitmapAlpha(thePicture, 0, 0, FillColor, msManagedBitmap, BmpAlphaUsage, 0, False)
        End If
    Case vbPicTypeIcon
        Set tPic = pvProcessHICON(thePicture, 0, 0, FillColor, msManagedBitmap, BmpAlphaUsage, 0)
    End Select
    
exitRoutine:
    If tPic Is Nothing Then Set tPic = thePicture
    Set ManagePicture = tPic

End Function

Public Function UnmanagePicture(thePicture As StdPicture, _
                    Optional ByVal FillColor As Long = -1, _
                    Optional ByVal ApplyExistingAttributes As Boolean = False) As StdPicture
                        
    ' Set Image1.Picture = thisClass.UnmanagePicture(Image1.Picture, Me.BackColor)
    ' Image1.Refresh
    
    ' To unmanage all pictures at once, set your cached oStdPictureEx object
    '   to Nothing and then set this class to Nothing to release its reference
    '   on your object. When that object is released, all management is released.
    
    '   FillColor: optional opaque color to replace transparency, VB system colors accepted
    '       -1 = will be black for unmanaged pictures having transparency.
    '       Otherwise, pass valid color to replace the transparency within the image.
    '   ApplyExistingAttributes
    '       If True, any attributes cached with GdipImageAttributes property are
    '       rendered into the picture before it is unmanaged.
    
    ' Unmanages a managed picture and returns the result
    ' Any image with transparency will have the transparency rendered as black.
    ' It is not recommended to unmanage a picture once it is managed.
    
    Dim hThunk As Long, lClientAddr As Long, hAttrs As Long
    Dim hClients As Long, hAddr As Long, oPic As StdPicture
    
    ' note: canary not checked nor is creation of GDI+ token
    ' if picture is managed, a canary does exist somewhere so does a GDI+ token
    hAddr = pvIsManaged(thePicture, lClientAddr, , hThunk, hClients)
    If hAddr <> 0 Then      ' else not managed
        If ApplyExistingAttributes = True Then
            ' no need to manually track, attrs will be disposed on CallThunkProc below
            CopyMemory hAttrs, ByVal ((hAddr Xor &H80000000) + Offset_ClientPropB) Xor &H80000000, 4
        End If
        If FillColor <> -1 Or hAttrs <> 0 Then
            If FillColor < 0 Then FillColor = GetSysColor(FillColor And &HFF)
            If thePicture.Type = vbPicTypeBitmap Then
                Set oPic = pvProcessBitmap(thePicture, 0, 0, FillColor, msUnmanagedBitmap, alphaNone, hAttrs, False)
            Else
                Set oPic = pvProcessHICON(thePicture, 0, 0, FillColor, msUnmanagedBitmap, acApplyAlphaChannel, hAttrs)
            End If
        End If
        CallThunkProc ((hThunk Xor &H80000000) + Offset_Unhook) Xor &H80000000, hThunk, hClients, lClientAddr, 0
    End If
    
exitRoutine:
    If oPic Is Nothing Then Set oPic = thePicture
    Set UnmanagePicture = oPic

End Function

Public Property Get IsPictureManaged(thePicture As StdPicture) As Boolean
    ' returns whether or not the passed Picture is managed/hooked
    If pvIsPictureValid(thePicture) = True Then
        IsPictureManaged = Not (pvGetCanary(thePicture) Is Nothing)
    End If
End Property

Public Property Get GdipImageAttributes(thePicture As StdPicture) As Long

    ' You are responsible for disposing of any returned attributes handles
    '    Use: GdipDisposeImageAttributes or class DeleteGDIpAttributesHandle
    ' Property returns the active handle for the passed managed picture
    
    Dim hAttrs As Long, gAttrs As Long
    hAttrs = pvSetImageAttributes(thePicture, 0, False)
    If hAttrs <> 0 Then
        GdipCloneImageAttributes hAttrs, gAttrs
        GdipImageAttributes = gAttrs
    End If

End Property
Public Property Let GdipImageAttributes(thePicture As StdPicture, ByVal hAttrs As Long)

    ' example after creating attributes:
    '   thisClass.GdipImageAttributes(Image1.Picture) = hAttributes
    '   GdipDisposeImage hAttributes ' must be disposed at some time
    '   Image1.Refresh
    
    ' You are responsible for disposing of the passed attributes handle
    '    Use: GdipDisposeImageAttributes or class DeleteGDIpAttributesHandle
    ' Property sets/removes the active handle for the passed managed picture
    ' Pass zero to remove the attributes or a valid handle to replace existing one
    
    ' GDI+ Image Attributes are a handy way to create grayscaling and global
    ' transparency. There are other things the attributes can do also
    
    pvSetImageAttributes thePicture, hAttrs, True

End Property

Public Property Get GdipScalingQuality(thePicture As StdPicture) As ScalingQuality

    ' Property returns the scaling quality for a managed picture
    ' Bicubic scaling is default for all managed pictures
    ' -1 returned if picture is unmanaged
    Dim hAddr As Long, lFlags As Long
    hAddr = pvIsManaged(thePicture)
    If hAddr = 0 Then
        GdipScalingQuality = -1
    Else
        CopyMemory lFlags, ByVal ((hAddr Xor &H80000000) + Offset_ClientPropA) Xor &H80000000, 4
        GdipScalingQuality = (lFlags And &H30000000) \ &H10000000
    End If

End Property
Public Property Let GdipScalingQuality(thePicture As StdPicture, ByVal lValue As ScalingQuality)
    
    ' Property sets the scaling quality for a managed picture.
    '   thisClass.GdipScalingQuality(Image1.Picture) = sqBiLinear
    '   Image1.Refresh
    
    ' see pvHookPicture for details of the client data array
    
    If lValue >= sqNearestNeighbor And lValue <= sqBiCubic Then
        Dim hAddr As Long, lFlags As Long
        hAddr = pvIsManaged(thePicture)
        If hAddr <> 0 Then
            CopyMemory lFlags, ByVal ((hAddr Xor &H80000000) + Offset_ClientPropA) Xor &H80000000, 4
            lFlags = (lFlags And &HCFFFFFFF) Or lValue * &H10000000
            CopyMemory ByVal ((hAddr Xor &H80000000) + Offset_ClientPropA) Xor &H80000000, lFlags, 4
        End If
    End If
End Property

Public Property Get GetGDIpHandle(thePicture As StdPicture) As Long

    ' returns the GDI+ handle (if any) used to manage the passed image
    ' DO NOT DESTROY THE RETURNED HANDLE - can cause thunk to crash
    ' DO NOT CACHE IT - can be dynamically replaced or destroyed within this class
    '   if calling this routine, use the handle immediately & then forget about it
    
    ' The handle can be used with any applicable GDI+ functions.
    ' All managed pictures have a GDI+ handle that is tied/linked to its VB picture
    
    pvIsManaged thePicture, , GetGDIpHandle

End Property

'///// end of public thunk/management related methods
#End If

Public Function LoadPictureEx(ByVal Source As Variant, _
                    Optional ByVal Width As Long = 0, Optional ByVal Height As Long = 0, _
                    Optional ByRef ImageIndex As Long = 0, Optional ByVal ImageDepth As Long = 0, _
                    Optional ByVal ManageStyle As ManagedStyle = msAutoSelect, _
                    Optional ByVal FillColor As Long = -1, _
                    Optional ByVal BmpAlphaUsage As AlphaUsageEnum = acApplyAlphaChannel, _
                    Optional ByVal hGdipAttributes As Long = 0) As StdPicture

    ' Similar to VB's LoadPicture but with several enhancements
    ' Set Image1.Picture = thisClass.LoadPictureEx("c:\temp\sample.png")
    
    ' If bitmap original depth must be returned, pass ManageStyle=acIgnoreAlphaChannel
    '   and all other parameters as defaults
    ' Comments below also apply, in large part, to LoadResPictureEx & CopyPicture
    
    '   Source: path/file to load, unicode is accepted. PNG/TIFF accepted
    '           a one dimensional byte array having entire image format
    '   Width,Height: optional end-result dimensions.
    '       zero will use actual size, negative values mirror
    '   ManageStyle: managing requires the class Attach method to be called
    '     msAutoSelect (default)
    '       icon/cursor: always return HICON,HCURSOR
    '       WMF/EMF: return VB loaded metafile
    '       GIF: VB loaded or navigation enabled depending on ImageIndex parameter
    '       All others: set to msUnmanaged if no transparency exists in image
    '           else set to msManaged if no FillColor has been passed.
    '       !! Except icon/cursor, passing non-default FillColor = msUnmanaged
    '     msUnmanagedBitmap: converts any format to bitmap, VB renders/scales
    '       if transparency exists and FillColor=-1, transparency rendered black
    '     msManagedBitmap: converts any format to bitmap
    '       GDI+ renders/scales the image as needed
    '       Thunk rendering applies only when the picture's Render method is called.
    '       That method is only called for pictures in the Form/Image/Picturebox
    '       picture property. Render is not called for any other picture properties
    '       nor during VB's PaintPicture method. When managed, avoid APIs like
    '       BitBlt/StretchBlt. Instead, use the managed picture's Render method.
    '   FillColor: optional opaque color to replace transparency, VB system colors accepted
    '       -1 (default) = allow transparency, will be black for unmanaged pictures.
    '       Otherwise, pass valid color to replace the transparency within the image.
    '   ImageIndex (ByRef)
    '       for icons/cursors, see chart below
    '       for TIFF: specify page to return: 0,1 = page #1, 2 = Page #2, etc
    '       for GIF: 0 (default) = 1st frame, no frame navigation
    '                1 = 1st frame, frame navigation enabled
    '               >1 = selected frame, no frame navigation
    '               See pvLoadFormatGIF for far more details
    '       returned value is count of frames/pages, icons/cursors
    '   ImageDepth: applies only to icons/cursors, see chart below
    '   AlphaUsage: ignored when returning icons/cursors
    '      acApplyAlphaChannel (default) = test for alpha channel & apply if it exists
    '      acIgnoreAlphaChannel = alpha not used. Set this for VB Clipboard pictures
    '           acIgnoreAlphaChannel can be combined with color keys
    '      acWantFormat32bppPARGB = return PARGB format if transparency exists
    '           acWantFormat32bppPARGB can be combined with color keys
    '      acColorKeyIsTopLeftPixel = all pixels same as top/left pixel color are transparent
    '      acColorKeyIsBlack = color black is transparent
    '      else any valid color = make transparent all pixels equal to this color
    '   hGdipAttributes: GDI+ image attributes handle, you are responsible for disposing
    '       Attributes are rendered into the passed picture
    '       Conversion to bitmap will be performed. Exception: Icon/Cursor & ManageStyle=msAutoSelect
    '       If attributes will introduce transparency and ManageStyle=msUnmanagedBitmap,
    '           always supply a FillColor to blend the image into
    '(*) note that if no thunk is created, any managed pictures that have transparency,
    '    will have transparency rendered as black without FillColor. Call the Attach method.

    ' Tips for returning icons...
    ' If ManageStyle = msAutoSelect, HICON/HCURSOR returned else HBITMAP returned
    ' Goal                                  ImageIndex  Width,Height   ImageDepth
    ' --------------------------------------------------------------------------
    ' Same as VB's LoadPicture* (32x32)(default) 0          0*          0
    ' Best image at system small size* (16x16)  -1          0*          0
    ' System size at specific depth              0 or -1    0*          depth
    ' Best image at specific size                0          size        0
    ' Best image at specific size & depth        0          size        depth
    ' Specific index, actual size                index      0           ignored
    ' Specific index, forced size                index      size        ignored
    ' (*) These are system DPI-aware since they use system metrics to retrieve the
    '     size, if your project is manifested as DPI-aware. 32x32, 16x16 @ 96 DPI
    '     Any other scenario requires you to pass Width,Height scaled, if desired
    ' (^) Black/White cursors are scaled using CopyImage API, no rendering occurs
    ' GDI+ is used to scale icons. The returned icon is always 32bpp
    '   If that is not acceptable, you should use CopyImage or other APIs.
    
    ' --------------------------------------------------------------------------
    ' Bitmap, PNG, TIFF, JPEG
    '   If AlphaUsage<>acIgnoreAlphaChannel & FillColor=-1 & ManageStyle=msAutoSelect
    '       then ManageStyle set to msManagedBitmap if transparency applies
    '   For TIFF, ImageIndex returns number of pages found in the TIFF
    '   For JPEG: auto-correction for orientation performed, if applies
    ' --------------------------------------------------------------------------
    ' Metafile
    '   If defaults are passed, VB loaded metafile returned
    '   Any non-default values results in conversion to bitmap
    ' --------------------------------------------------------------------------
    ' GIF
    '   If defaults are passed, VB loaded GIF returned, no frame navigation
    '   elseif ImageIndex=1 & defaults for: ManageStyle,FillColor,AlphaUsage,hGdipAttributes
    '       then returned GIF frames can be navigated via picture's hPal property
    '   else conversion to bitmap is performed and no frame navigation
    '   ImageIndex returns number frames in the GIF
    ' --------------------------------------------------------------------------

    Dim aData() As Byte, oStream As stdole.IUnknown
    
    If pvCreateToken(0&, True) = False Then Exit Function
    If VarType(Source) = vbString Then
        If LenB(Source) = 0 Then Exit Function
        If pvGetFileData(Source, aData()) = False Then Exit Function
    ElseIf VarType(Source) = (vbArray Or vbByte) Then
        aData() = Source
    Else
        Exit Function
    End If
    
    If ManageStyle < msAutoSelect Then
        ManageStyle = msAutoSelect
    ElseIf ManageStyle > msManagedBitmap Then
        ManageStyle = msManagedBitmap
    End If
    If FillColor <> -1 Then
        If FillColor < 0 Then FillColor = GetSysColor(FillColor And &HFF)
    End If
    BmpAlphaUsage = BmpAlphaUsage And AlphaUsageMask
    If (BmpAlphaUsage And (acIgnoreAlphaChannel Or acWantFormat32bppPARGB)) = (acIgnoreAlphaChannel Or acWantFormat32bppPARGB) Then
        BmpAlphaUsage = BmpAlphaUsage Xor acWantFormat32bppPARGB
    End If
    
    Set LoadPictureEx = pvLoadPicture(aData(), Width, Height, ManageStyle, _
                        FillColor, ImageIndex, ImageDepth, BmpAlphaUsage, hGdipAttributes)
'    If LoadPictureEx Is Nothing Then Err.Raise 481
    
End Function

Public Function LoadResPictureEx(ResID As Variant, ResType As Variant, _
                    Optional ByVal Width As Long = 0, Optional ByVal Height As Long = 0, _
                    Optional ByRef ImageIndex As Long = 0, Optional ByVal ImageDepth As Long = 0, _
                    Optional ByVal ManageStyle As ManagedStyle = msAutoSelect, _
                    Optional ByVal FillColor As Long = -1, _
                    Optional ByVal BmpAlphaUsage As AlphaUsageEnum = acApplyAlphaChannel, _
                    Optional ByVal hGdipAttributes As Long = 0) As StdPicture
                    
    ' Similar to VB's LoadResPicture but with several enhancements
    '   i.e., supports PNG, TIFF formats in custom resource sections
    
    ' Set Image1.Picture = thisClass.LoadResPictureEx(101, vbResBitmap)
    ' Set Image2.Picture = thisClass.LoadResPictureEx(101, "CUSTOM")
    
    ' ResID is the resource file ID/Key for the picture/data, i.e., 101
    ' ResType is a numeric LoadResConstants value or custom name, i.e., "CUSTOM"
    ' For other parameters' details, see LoadPictureEx
    
    If IsEmpty(ResID) = True Or IsEmpty(ResType) = True Then Exit Function
    
    Dim tPic As StdPicture, aData() As Byte, icCount As Long
    
    Const RT_CURSOR As Long = 1, RT_ICON As Long = 3
    
    If pvCreateToken(0&, True) = False Then Exit Function
    
    If ManageStyle < msAutoSelect Then
        ManageStyle = msAutoSelect
    ElseIf ManageStyle > msManagedBitmap Then
        ManageStyle = msManagedBitmap
    End If
    If FillColor <> -1 Then
        If FillColor < 0 Then FillColor = GetSysColor(FillColor And &HFF)
    End If
    BmpAlphaUsage = BmpAlphaUsage And AlphaUsageMask
    If (BmpAlphaUsage And (acIgnoreAlphaChannel Or acWantFormat32bppPARGB)) = (acIgnoreAlphaChannel Or acWantFormat32bppPARGB) Then
        BmpAlphaUsage = BmpAlphaUsage Xor acWantFormat32bppPARGB
    End If
    
    If VarType(ResType) = vbString Then ' custom resources
        On Error GoTo exitRoutine
        aData() = VB.LoadResData(ResID, ResType)
        On Error GoTo 0
    
    ElseIf ResType = vbResBitmap Then
        ' note: bitmaps loaded via LoadResPicture are DDB vs DIB
        ' If DIB is required, add it to custom resource instead
        On Error GoTo exitRoutine
        Set tPic = VB.LoadResPicture(ResID, ResType)
        On Error GoTo 0
        If pvIsPictureValid(tPic) Then
            Set LoadResPictureEx = pvProcessBitmapAlpha(tPic, Width, Height, FillColor, ManageStyle, BmpAlphaUsage, hGdipAttributes, True)
        End If
        GoTo exitRoutine
    
    ElseIf ResType = vbResCursor Or ResType = vbResIcon Then
        ' this portion reconstructs the resource to include all its icons
        '   into an array that is processed by pvLoadFormatIcon. if
        '   loading by passed ImageIndex, then only that resource is added
        '   to the reconstruction.
        ' The main difference between res Directory structures and those
        '   in files are size and icon ID. Res structures are 14 bytes and
        '   16 bytes in files. Res structures have 2-byte icon ID for location
        '   of the icon data whereas files have a 4-byte offset into the file.
        Dim arrDir() As Byte, arrRes() As Byte
        Dim lCount As Long, n As Long, lPos As Long
        Dim lSize As Long, lOffset As Long
        Dim lType As Long, lValue As Long
        
        If ResType = vbResIcon Then               ' icons/cursors are each different within res data
            lType = RT_ICON
        Else
            lType = RT_CURSOR: lOffset = 4        ' cursors contain a 4 byte prefix for hotspot
        End If
        
        On Error GoTo exitRoutine
        arrDir() = VB.LoadResData(ResID, lType + 11) ' get the directory for the passed res icon/cursor
        For n = 14 To UBound(arrDir) - 5 Step 14  ' count number of sub-images & total size of all sub-images
            CopyMemory lValue, arrDir(n), 4       ' size of resource, every 14 bytes from offset 14
            lSize = lSize + lValue - lOffset      ' cur: includes hotspot, remove from size
            lCount = lCount + 1: icCount = lCount
            If lCount = ImageIndex Then           ' abort if loading icon by index
                lSize = lValue: lCount = 1
                n = -n: Exit For
            End If
        Next
        
        ReDim aData(0 To lCount * 16 + lSize + 5) ' size array for all sub-images + directories
        ' lPos = arrData slot, lOffset = location in arrData for image bytes
        lPos = 6: lOffset = lCount * 16 + lPos
        CopyMemory aData(0), arrDir(0), lPos      ' copy 1st 6 bytes from directory
        If n < 0 Then                             ' either just 1 image or loading by index
            n = -n + 4: lValue = n                ' set loop start/stop on icon IDs
            aData(4) = 1: aData(5) = 0            ' and ensure count is correct
        Else
            n = 18: lValue = UBound(arrDir) - 1   ' set loop start/stop on icon IDs
        End If
        For n = n To lValue Step 14 ' loop thru each directory & locate the res icon/cursor ID
            lValue = arrDir(n) Or arrDir(n + 1) * &H100&    ' res file's iconID
            arrRes() = VB.LoadResData(lValue, lType) ' load that resource now & cache its size
            lSize = UBound(arrRes) + 1&
            If lType = RT_CURSOR Then             ' build the icon/cursor directory
                lSize = lSize - 4: CopyMemory aData(lOffset), arrRes(4), lSize
            Else
                CopyMemory aData(lOffset), arrRes(0), lSize
            End If
            CopyMemory aData(lPos), arrDir(n - 12), 8  ' copy 1st 8 bytes of directory entry
            CopyMemory aData(lPos + 8), lSize, 4       ' add resource size
            CopyMemory aData(lPos + 12), lOffset, 4    ' add location to directory
            
            lPos = lPos + 16                      ' set next directory location
            lOffset = lOffset + lSize             ' set next slot in our array for res data
        Next
        Erase arrDir(): Erase arrRes()
        On Error GoTo 0
    Else
        GoTo exitRoutine
    End If
    
    Set tPic = pvLoadPicture(aData(), Width, Height, ManageStyle, _
                        FillColor, ImageIndex, ImageDepth, BmpAlphaUsage, hGdipAttributes)
    If icCount <> 0 Then ImageIndex = icCount
    Set LoadResPictureEx = tPic
    
exitRoutine:
    If Err Then Err.Clear
'    If LoadResPictureEx Is Nothing Then Err.Raise 481
    
End Function

Public Function CopyPicture(thePicture As IPicture, _
                        Optional ByVal Width As Long = 0, Optional ByVal Height As Long, _
                        Optional ByVal ScaleFactor As Single = 1, _
                        Optional ByVal ManageStyle As ManagedStyle = msAutoSelect, _
                        Optional ByVal FillColor As Long = -1, _
                        Optional ByVal BmpAlphaUsage As AlphaUsageEnum = acApplyAlphaChannel, _
                        Optional ByVal hGdipAttributes As Long = 0) As StdPicture
                        
    ' Set Image1.Picture = thisClass.CopyPicture(Image1.Picture, 256, 256)
    ' For example, if DPI increase/decreases by 25%:
    '   Set Image1.Picture = thisClass.CopyPicture(Image1.Picture, , , 1.25)
    '   Set Image1.Picture = thisClass.CopyPicture(Image1.Picture, , , 0.75)
    ' Example of using cached attributes to render into the copy...
    '   Dim hAttrs As Long
    '   hAttrs = thisClass.GdipImageAttributes(Image1.Picture)
    '   Set Image1.Picture = thisClass.CopyPicture(Image1.Picture, , , 1.25, , , , hAttrs)
    '   thisClass.DeleteGDIpAttributesHandle hAttrs
                        
    ' Returns a new picture object based on parameters & passed picture, with
    '   only exception being navigation-enabled GIFs (see next paragraph)
    ' All parameters apply to the returned picture, not the passed picture
    
    ' To scale a GIF with frame navigation enabled, all parameters except Width,
    '   Height, or ScaleFactor must be default values. Any other scenario, and
    '   a copy of the current GIF frame is returned. The source is not scaled.
    '   If no change in size occurs, the same picture object is returned. To
    '   return a copy at the same size for the current frame, pass ManageStyle
    '   as either msManagedBitmap or msUnmanagedBitmap.
    
    ' Width,Height: new image dimensions, zero = unchanged
    '   negative values result in mirroring. if mirroring, generally best
    '   to ensure picture's mirrored dimensions are odd, not even before mirroring.
    ' ScaleFactor: prevents having to supply new Width,Height if scaling actual
    '   size to a known scale factor. When passed and > 0, image actual size will
    '   be scaled to the passed factor. If Width or Height is non-zero, then
    '   this parameter is ignored. Setting to 1 is same as passing Width,Height as 0
    ' ManageStyle: managing requires the class Attach method to be called
    '   msAutoSelect (default)
    '       icon/cursor: always return HICON,HCURSOR
    '       GIF: see notes above
    '       All others: existing format converted to bitmap.
    '           sets to msUnmanaged if no transparency exists in image
    '           else sets to msManaged if FillColor = -1 or hGdipAttributes passed.
    '       !! Except icon/cursor, passing non-default FillColor = msUnamnaged
    '       !! Except icon/cursor, passing non-zero hGdipAttributes = msManaged
    '   msUnmanagedBitmap: converts any format to bitmap, VB renders/scales
    '       if transparency exists and FillColor=-1, transparency rendered black
    '   msManagedBitmap: converts any format to bitmap (alpha-supported)
    '       GDI+ renders/scales the image as needed
    '       Thunk rendering applies only when the picture's Render method is called.
    '       That method is only called for pictures in the Form/Image/Picturebox
    '       picture property. Render is not called for any other picture properties
    '       nor during VB's PaintPicture method. When managed, avoid APIs like
    '       BitBlt/StretchBlt. Instead, use the managed picture's Render method.
    ' FillColor: optional opaque color to replace transparency, VB system colors accepted
    '   -1 (default) = allow transparency, will be black for unmanaged pictures.
    '   Otherwise, pass valid color to replace the transparency within the image.
    ' AlphaUsage: ignored when returning icons/cursors
    '   acApplyAlphaChannel (default) = test for alpha channel & apply if it exists
    '   acIgnoreAlphaChannel = alpha not used. Set this for VB Clipboard pictures
    '       acIgnoreAlphaChannel can be combined with color keys
    '   acWantFormat32bppPARGB = return PARGB format if transparency exists
    '       acWantFormat32bppPARGB can be combined with color keys
    '   acColorKeyIsTopLeftPixel = all pixels same as top/left pixel color are transparent
    '   acColorKeyIsBlack = color black is transparent
    '   else any valid color = make transparent all pixels equal to this color
    ' hGdipAttributes: GDI+ image attributes handle, you are responsible for disposing
    '   Attributes are rendered into the passed picture
    '   Conversion to bitmap will be performed. Exception: Icon/Cursor & ManageStyle=msAutoSelect
    '   If attributes will introduce transparency and ManageStyle=msUnmanagedBitmap,
    '       always supply a FillColor to blend the image into
    '(*) note that if no thunk is created, any managed pictures that have transparency,
    '    will have transparency rendered as black without FillColor. Call the Attach method.
    
    If pvIsPictureValid(thePicture) = False Then Exit Function
    
    Dim lValue As Long, Cx As Long, Cy As Long
    Dim oPic As StdPicture
    
    If pvCreateToken(0&, True) = False Then Exit Function
    If ManageStyle < msAutoSelect Then
        ManageStyle = msAutoSelect
    ElseIf ManageStyle > msManagedBitmap Then
        ManageStyle = msManagedBitmap
    End If
    If FillColor <> -1 Then
        If FillColor < 0 Then FillColor = GetSysColor(FillColor And &HFF)
    End If
    If (Width = 0 Or Height = 0) And ScaleFactor <> 1 Then
        If ScaleFactor > 0 Then
            Me.GetImageDimensions thePicture, Cx, Cy
            If Width = 0 Then Width = Cx * ScaleFactor
            If Height = 0 Then Height = Cy * ScaleFactor
        End If
    End If
    BmpAlphaUsage = BmpAlphaUsage And AlphaUsageMask
    If (BmpAlphaUsage And (acIgnoreAlphaChannel Or acWantFormat32bppPARGB)) = (acIgnoreAlphaChannel Or acWantFormat32bppPARGB) Then
        BmpAlphaUsage = BmpAlphaUsage Xor acWantFormat32bppPARGB
    End If
    
    Select Case thePicture.Type
    Case vbPicTypeEMetafile, vbPicTypeMetafile
        Set oPic = pvConvertMetafile(thePicture, Width, Height, FillColor, ManageStyle, BmpAlphaUsage, hGdipAttributes)
        
    Case vbPicTypeBitmap
        If (thePicture.Attributes And PICTURE_TRANSPARENT) = PICTURE_TRANSPARENT Then
            lValue = pvIsManaged(thePicture)
            If lValue = 0 Then  ' VB GIF
                Set oPic = pvConvertGIF(thePicture, Width, Height, FillColor, ManageStyle, BmpAlphaUsage, hGdipAttributes)
                lValue = 1
            Else
                #If AllowThunkCapability = True Then
                CopyMemory lValue, ByVal ((lValue Xor &H80000000) + Offset_ClientPropA) Xor &H80000000, 4
                lValue = lValue And &H80000000  ' GIF with frame-navigation enabled
                If lValue <> 0 Then
                    Set oPic = pvScaleAnimatedGIF(thePicture, Width, Height, FillColor, ManageStyle, BmpAlphaUsage, hGdipAttributes, False)
                End If
                #End If
            End If
        End If
        If lValue = 0 Then
            Set oPic = pvProcessBitmapAlpha(thePicture, Width, Height, FillColor, ManageStyle, BmpAlphaUsage, hGdipAttributes, False)
        End If
        
    Case vbPicTypeIcon
        Set oPic = pvProcessHICON(thePicture, Width, Height, FillColor, ManageStyle, BmpAlphaUsage, hGdipAttributes)
    End Select
    
exitRoutine:
    If oPic Is thePicture Then  ' don't return same picture unless navigation-enabled GIF
        If lValue <> &H80000000 Then Set oPic = Nothing
    End If
    Set CopyPicture = oPic

End Function

Public Function CopyPictureToHandle(thePicture As StdPicture, ByVal AsBitmap As Boolean) As Long

    ' function will take passed picture and return a bitmap or icon handle
    ' You are responsible for disposing of the handle
    
    ' AsBitmap: True = returned bitmap handle compatible with pre-multipled RGB
    '           False = returned icon handle (CopyImage API if source is also icon)
    
    If pvIsPictureValid(thePicture) = False Then Exit Function
    
    Dim oPic As IPicture, lValue As Long
    Dim hHandle As Long, fRect() As Single
    Dim gHandle As Long, aSource() As Long
    Const LR_CREATEDIBSECTION As Long = &H2000
    
    If pvCreateToken(0&, True) = False Then Exit Function
    
    Select Case thePicture.Type
    Case vbPicTypeBitmap
        Set oPic = thePicture
        If pvIsManaged(thePicture, , gHandle) = 0 Then
            If (oPic.Attributes And PICTURE_TRANSPARENT) = PICTURE_TRANSPARENT Then ' unmanaged GIF
                Set oPic = pvConvertGIF(thePicture, 0, 0, -1, msUnmanagedBitmap, acApplyAlphaChannel, 0)
            ElseIf AsBitmap = True Then
                ReDim aSource(0 To 5)
                GetGDIObject oPic.Handle, 24, aSource(0)
                If (aSource(4) And &HFF0000) = &H200000 And aSource(5) <> 0 Then
                    ' convert 32bpp ARGB to pARGB as needed
                    If pvValidateAlphaChannel(aSource(5), aSource(1) * aSource(2)) = PixelFormat32bppARGB Then
                        lValue = aSource(5)
                        aSource(5) = pvCreateDibDC(hHandle, aSource(1), aSource(2), 32)
                        If aSource(5) = 0 Then GoTo exitRoutine
                        GdipCreateBitmapFromScan0 aSource(1), aSource(2), aSource(3), PixelFormat32bppARGB, ByVal lValue, gHandle
                        pvConvertFormat gHandle, PixelFormat32bppPARGB, aSource(5), aSource(3)
                        GdipDisposeImage gHandle: gHandle = 0
                        GoTo exitRoutine
                    End If
                End If
                Erase aSource()
            End If
        End If
    Case vbPicTypeMetafile, vbPicTypeEMetafile
        Set oPic = pvConvertMetafile(thePicture, 0, 0, -1, msUnmanagedBitmap, acApplyAlphaChannel, 0)
    Case vbPicTypeIcon
        If AsBitmap = False Then
            hHandle = CopyImage(thePicture.Handle, 1, 0, 0, 0)
            GoTo exitRoutine
        Else
            Set oPic = pvProcessHICON(thePicture, 0, 0, -1, msUnmanagedBitmap, acApplyAlphaChannel, 0)
        End If
    Case Else
    End Select
    If Not oPic Is Nothing Then
        If AsBitmap = False Then
            pvProcessHICON oPic, 0, 0, -1, msAutoSelect, acApplyAlphaChannel, 0, True, hHandle
        ElseIf oPic Is thePicture Then  ' unmanged non-GIF or managed bitmap
            If gHandle <> 0 Then
                GdipGetImagePixelFormat gHandle, lValue
                If lValue = PixelFormat32bppARGB Then
                    ReDim fRect(0 To 3)
                    GdipGetImageBounds gHandle, fRect(0), lValue
                    hHandle = pvDIBFromGDIplus(gHandle, (fRect(2)), (fRect(3)), -1, , , fRect())
                    Erase fRect()
                End If
            End If
        End If
        If hHandle = 0 And AsBitmap = True Then
            hHandle = CopyImage(oPic.Handle, 0, 0, 0, LR_CREATEDIBSECTION)
        End If
    End If

exitRoutine:
    Set oPic = Nothing
    CopyPictureToHandle = hHandle

End Function

Public Function CreateGDIpAttributesHandle(Optional ByVal GrayScaleType As ColorScaleRatioEnum = gsNone, _
                                Optional ByVal GlobalOpacityPercent As Byte = 100, _
                                Optional ByVal InvertedColors As Boolean = False, _
                                Optional ByVal UseWrapModeTileFlipXY As Boolean = True) As Long
                                        
    ' Courtesy function, not used in this class. Creates a GDI+ image attributes handle
    ' You are reponsible for disposing of any returned handles
    '    Use: GdipDisposeImageAttributes or class DeleteGDIpAttributesHandle
    
    ' FYI: UseWrapModeTileFlipXY helps prevent blending of outer edges when an
    '   opaque image is scaled with high quality
    
    '--------------------------------------------------------------------------------
    '   WARNING     WARNING     WARNING     WARNING     WARNING     WARNING
    '--------------------------------------------------------------------------------
    ' Any handle returned here must be released before one of two things occur
    '   1. The class unloads (which releases its token)
    '   2. A GDI+ token you may already have is released
    ' Any GDI+ objects are invalidated when the last GDI+ token is released
    
    Dim Col As Long, Row As Long, hAttrs As Long
    Dim clrMatrix(0 To 4, 0 To 4) As Single
    Const ColorAdjustTypeBitmap As Long = &H1
    
    If pvCreateToken(0&, True) = False Then Exit Function
    If GrayScaleType Then
        ' grayscale the image; populate 1st column, 1st 3 rows
        Select Case GrayScaleType
            Case gsNTSCPAL ' standard weighted average
                clrMatrix(Col, Row) = 0.299: clrMatrix(Col, Row + 1) = 0.587: clrMatrix(Col, Row + 2) = 0.114
            Case gsCCIR709 ' CCIR709
                clrMatrix(Col, Row) = 0.213: clrMatrix(Col, Row + 1) = 0.715: clrMatrix(Col, Row + 2) = 0.072
            Case gsSimpleAverage ' pure average
                clrMatrix(Col, Row) = 0.333: clrMatrix(Col, Row + 1) = 0.334: clrMatrix(Col, Row + 2) = clrMatrix(Col, Row)
            Case gsRedMask ' personal preferences: could be r=1:g=0:b=0 or other weights
                clrMatrix(Col, Row) = 0.8: clrMatrix(Col, Row + 1) = 0.1: clrMatrix(Col, Row + 2) = clrMatrix(Col, Row + 1)
            Case gsGreenMask ' personal preferences: could be r=0:g=1:b=0 or other weights
                clrMatrix(Col, Row) = 0.1: clrMatrix(Col, Row + 1) = 0.8: clrMatrix(Col, Row + 2) = clrMatrix(Col, Row)
            Case gsBlueMask ' personal preferences: could be r=0:g=0:b=1 or other weights
                clrMatrix(Col, Row) = 0.1: clrMatrix(Col, Row + 1) = clrMatrix(Col, Row): clrMatrix(Col, Row + 2) = 0.8
            Case gsRedGreenMask ' personal preferences: could be r=.5:g=.5:b=0 or other weights
                clrMatrix(Col, Row) = 0.45: clrMatrix(Col, Row + 1) = clrMatrix(Col, Row): clrMatrix(Col, Row + 2) = 0.1
            Case gsBlueGreenMask ' personal preferences: could be r=0:g=.5:b=.5 or other weights
                clrMatrix(Col, Row) = 0.1: clrMatrix(Col, Row + 1) = 0.45: clrMatrix(Col, Row + 2) = clrMatrix(Col, Row + 1)
            Case gsAlphaMask    ' pixel alpha value applied to R,G,B
                clrMatrix(0, 3) = 1: clrMatrix(1, 3) = 1: clrMatrix(2, 3) = 1
                clrMatrix(4, 3) = 1
            Case gsSepia ' populate 1st 3 columns, 1st 3 rows
                clrMatrix(0, 0) = 0.393: clrMatrix(1, 0) = 0.349: clrMatrix(2, 0) = 0.272
                clrMatrix(0, 1) = 0.769: clrMatrix(1, 1) = 0.686: clrMatrix(2, 1) = 0.534
                clrMatrix(0, 2) = 0.189: clrMatrix(1, 2) = 0.168: clrMatrix(2, 2) = 0.131
            Case Else ' no grayscale
                GrayScaleType = gsNone
        End Select
        If GrayScaleType <> gsNone Then
            clrMatrix(Col, Row + 4) = 0.001         ' add minor offset to prevent GDI+ overflow
            If GrayScaleType = gsSepia Then
                clrMatrix(Col + 1, Row + 4) = 0.001: clrMatrix(Col + 2, Row + 4) = 0.001
            Else ' fill in columns 2 & 3 of 1st 4 rows
                For Row = 0& To 4&
                    For Col = 1& To 2&
                        clrMatrix(Col, Row) = clrMatrix(0, Row)
                Next: Next
            End If
            clrMatrix(4, 4) = 1!                    ' flag indicating need to create attributes
        End If
    End If
    
    If (InvertedColors = True Or GlobalOpacityPercent < 100 Or UseWrapModeTileFlipXY = True) Then
        If clrMatrix(4, 4) = 0! Then
            clrMatrix(0, 0) = 1!: clrMatrix(1, 1) = 1!: clrMatrix(2, 2) = 1!
            clrMatrix(4, 4) = 1!                    ' flag indicating need to create attributes
        End If
        If InvertedColors = True Then
            For Col = 0& To 2&
                For Row = 0& To 2&
                    clrMatrix(Col, Row) = -clrMatrix(Col, Row)
                 Next
                clrMatrix(Col, 4) = -clrMatrix(Col, 4) + 1!
            Next
        End If
    End If
    
    If clrMatrix(4, 4) = 1! Then                    ' create attributes?
        ' create the image attribute class and set the color matrix
        If GdipCreateImageAttributes(hAttrs) = 0& Then
            ' global blending; value between 0 & 1
            If GlobalOpacityPercent > 100 Then GlobalOpacityPercent = 100
            clrMatrix(3, 3) = CSng(GlobalOpacityPercent / 100!)
            If GdipSetImageAttributesColorMatrix(hAttrs, ColorAdjustTypeBitmap, 1&, clrMatrix(0, 0), clrMatrix(0, 0), 0&) <> 0& Then
                GdipDisposeImageAttributes hAttrs
            ElseIf UseWrapModeTileFlipXY = True Then
                GdipSetImageAttributesWrapMode hAttrs, WrapModeTileFlipXY, 0, 0
            End If
        End If
    End If
    CreateGDIpAttributesHandle = hAttrs
                                
End Function

Public Sub DeleteGDIpAttributesHandle(ByVal hHandle As Long)
    ' courtesty method
    If hHandle <> 0 Then
        If pvCreateToken(0&, True) = True Then
            GdipDisposeImageAttributes hHandle
        End If
    End If
    
End Sub

Public Function GetAlphaFormat(thePicture As StdPicture) As AlphaTypeEnum
    ' courtesy function returns the alpha channel format
    
    Dim aBits() As Long, gHandle As Long, lValue As Long
    Dim aSource(0 To 5) As Long, aData() As Long
    
    If pvIsPictureValid(thePicture) = False Then Exit Function
    pvIsManaged thePicture, , gHandle
    If gHandle <> 0 Then
        If pvCreateToken(0&, True) = True Then
            GdipGetImagePixelFormat gHandle, lValue
            If lValue < alphaExists Then GoTo exitRoutine
        Else
            gHandle = 0
        End If
    End If
    GetGDIObject thePicture.Handle, 24, aSource(0)
    If aSource(4) = &H200001 Or gHandle <> 0 Then ' 32 bits or has GDI+ handle
        If aSource(5) = 0 Or gHandle <> 0 Then
            ReDim aBits(0 To aSource(1) * aSource(2) - 1)
            aSource(5) = VarPtr(aBits(0))
            If gHandle = 0 Then
                ReDim aData(0 To 9)
                aData(0) = 40: aData(1) = aSource(1): aData(2) = aSource(2): aData(3) = &H200001
                lValue = GetDC(0)
                GetDIBits lValue, thePicture.Handle, 0, aSource(2), aBits(0), aData(0), 0
                Erase aData(): ReleaseDC 0, lValue
            Else
                pvConvertFormat gHandle, lValue, aSource(5), aSource(1) * 4
            End If
        End If
        Select Case pvValidateAlphaChannel(aSource(5), aSource(1) * aSource(2))
            Case PixelFormat32bppARGB: GetAlphaFormat = alphaARGB
            Case PixelFormat32bppPARGB: GetAlphaFormat = alphaPARGB
            Case PixelFormat32bppUARGB: GetAlphaFormat = alphaUniversal
        End Select
    End If
exitRoutine:
    
End Function

Public Sub GetImageDimensions(thePicture As StdPicture, Width As Long, Height As Long)
    ' simple routine to return image dimensions relative to actual system DPI
    If pvIsPictureValid(thePicture) = True Then
        Width = thePicture.Width / 2540 * m_DPI
        Height = thePicture.Height / 2540 * m_DPI
    Else
        Width = 0: Height = 0
    End If
End Sub

Public Sub RenderPicture(thePicture As StdPicture, ByVal hDC As Long, _
                Optional ByVal DstX As Long, Optional ByVal DstY As Long, _
                Optional ByVal DstWidth As Long, Optional ByVal DstHeight As Long, _
                Optional ByVal SrcX As Long, Optional ByVal SrcY As Long, _
                Optional ByVal SrcWidth As Long, Optional ByVal SrcHeight As Long)
                
    ' can be used like PaintPicture but rendered to any hDC, for any picture object
    ' this should be called instead of PaintPicture for managed pictures
    '   PaintPicture uses BitBlt/StretchBlt which will not include transparency
    
    ' all coordinates and dimensions are passed as pixel values
    ' any dimensions passed as zero will use picture's actual dimensions
    '   negative Width or Height parameters results in mirroring
    
    ' note: if picture is managed and has cached image attributes handle,
    ' then those attributes are rendered into the destination
    
    If hDC = 0 Then Exit Sub                ' invalid parameter checks
    If SrcY < 0 Or SrcX < 0 Then Exit Sub
    If DstY < 0 Or DstX < 0 Then Exit Sub
    If pvIsPictureValid(thePicture) = False Then Exit Sub
    
    Dim xRatio As Single
    xRatio = 2540! / m_DPI  ' multiply=Pixel>Himetric; divide=Himetric>Pixel
    With thePicture
        If DstWidth = 0 Then DstWidth = .Width / xRatio
        If DstHeight = 0 Then DstHeight = .Height / xRatio
        If SrcX <> 0 Then SrcX = SrcX * xRatio
        If SrcY <> 0 Then SrcY = SrcY * xRatio
        If SrcWidth = 0 Then SrcWidth = .Width Else SrcWidth = SrcWidth * xRatio
        If SrcHeight = 0 Then SrcHeight = .Height Else SrcHeight = SrcHeight * xRatio
        ' handle mirroring offsets
        '   mirror destination dimensions, not source dimensions
        If SrcWidth < 0 Then SrcWidth = -SrcWidth: DstWidth = -DstWidth
        If SrcHeight < 0 Then SrcHeight = -SrcHeight: DstHeight = -DstHeight
        If DstWidth < 0 Then DstX = DstX - DstWidth
        If DstHeight < 0 Then DstY = DstY - DstHeight
        
        .Render (hDC), (DstX), (DstY), (DstWidth), (DstHeight), _
            (SrcX), (.Height - SrcY), (SrcWidth), (-SrcHeight), ByVal 0&
    End With

End Sub

Public Function WrapHandleToStdPicture(ByVal ImageHandle As Long) As StdPicture

    ' Set Image1.Picture = thisClass.WrapHandleToStdPicture(hBitmap)

    ' can be used to create a stdPicture object from a HBITMAP, HICON, or HCURSOR
    ' handle. Only pass handles from APIs that create bitmaps/icons/cursors.
    ' - do not pass any existing stdPicture.Handle to this routine.
    ' - do not pass GDI+ image handles, this method only applies to GDI, not GDI+
    
    ' The returned StdPicture is wrapped around the passed handle. Do not destroy
    ' the handle. The handle will be destroyed when the picture is set to nothing
    ' or goes out of scope.
    
    Dim aData() As Long
    
    If ImageHandle = 0 Then Exit Function
    ReDim aData(0 To 5)
    If GetGDIObject(ImageHandle, 24, aData(0)) <> 0 Then
        Set WrapHandleToStdPicture = pvHandleToPicture(ImageHandle, vbPicTypeBitmap)
    ElseIf GetIconInfo(ImageHandle, aData(0)) <> 0 Then
        If aData(4) <> 0 Then DeleteObject aData(4)
        If aData(3) <> 0 Then   ' all icons/cursors have a mask - if not, not valid
            DeleteObject aData(3)
            Set WrapHandleToStdPicture = pvHandleToPicture(ImageHandle, vbPicTypeIcon)
        End If
    End If
    
End Function

Private Function pvLoadPicture(aData() As Byte, _
                    ByVal Width As Long, ByVal Height As Long, _
                    ByVal mgmtStyle As ManagedStyle, ByVal FillColor As Long, _
                    ByRef ImageIndex As Long, ByVal ImageDepth As Long, _
                    ByVal BmpAlphaUsage As AlphaUsageEnum, _
                    ByVal hGdipAttributes As Long) As StdPicture

    ' called from LoadPictureEx & LoadResPictureEx

    Dim oStream As stdole.IUnknown
    Dim picType As ePicType
    
    picType = pvGuessImageType(aData())
    If Not (picType = typeIcon Or picType = typeCursor) Then
        Set oStream = pvStreamFromArray(aData())
        Erase aData()
        If oStream Is Nothing Then Exit Function
    End If
    
    Select Case picType
    Case typeBitmap
        Set pvLoadPicture = pvLoadFormatBitmap(oStream, Width, Height, FillColor, mgmtStyle, BmpAlphaUsage, hGdipAttributes)
    Case typeJPG
        Set pvLoadPicture = pvLoadFormatJPEG(oStream, Width, Height, FillColor, mgmtStyle, BmpAlphaUsage, hGdipAttributes)
    Case typeIcon, typeCursor
        Set pvLoadPicture = pvLoadFormatIcon(aData(), Width, Height, FillColor, mgmtStyle, ImageIndex, ImageDepth, BmpAlphaUsage, hGdipAttributes)
        Erase aData()
    Case typeGIF
        Set pvLoadPicture = pvLoadFormatGIF(oStream, Width, Height, FillColor, BmpAlphaUsage, mgmtStyle, ImageIndex, hGdipAttributes)
    Case typeMetafile
        Set pvLoadPicture = pvLoadFormatMetafile(oStream, Width, Height, mgmtStyle, FillColor, BmpAlphaUsage, hGdipAttributes)
    Case Else   ' PNG, TIFF, or if pvGuessImageType guessed wrong
        Set pvLoadPicture = pvLoadFormatGdiPlus(oStream, Width, Height, mgmtStyle, FillColor, BmpAlphaUsage, picType, hGdipAttributes, ImageIndex)
    End Select
    
exitRoutine:

End Function

Private Function pvLoadFormatMetafile(oStream As stdole.IUnknown, Width As Long, Height As Long, _
                            mgmtStyle As ManagedStyle, FillColor As Long, _
                            alphaUsage As AlphaUsageEnum, rndrAttrs As Long) As StdPicture

    ' VB is used to load metafiles unless to be filled or managed or sized
    ' GDI+ will be used if VB fails.

    Dim tPic As StdPicture
    
    If Width = 0 And Height = 0 And alphaUsage = acApplyAlphaChannel Then
        If mgmtStyle = msAutoSelect And FillColor = -1 And rndrAttrs = 0 Then
            Set tPic = pvStreamToPicture(oStream)
        End If
    End If
    If tPic Is Nothing Then
        Set tPic = pvLoadFormatGdiPlus(oStream, Width, Height, mgmtStyle, FillColor, alphaUsage, typeMetafile, rndrAttrs)
    End If
    Set pvLoadFormatMetafile = tPic

End Function

Private Function pvLoadFormatGIF(oStream As stdole.IUnknown, _
                            Width As Long, Height As Long, FillColor As Long, _
                            alphaUsage As AlphaUsageEnum, _
                            mgmtStyle As ManagedStyle, ImageIndex As Long, _
                            rndrAttrs As Long) As StdPicture

    ' Uses VB to load GIF only when ImageIndex is passed as 0
    ' Otherwise, GDI+ is used
    
    ' ImageIndex
    ' 0 = VB used & falls back to GDI+ if fails. No frame navigation supported
    '     Exception: If to be filled or managed, then GDI+ is used.
    ' 1 = GDI+ used and frame navigation supported if more than 1 frame
    '     Exception: If FillColor & alphaUsage not default, navigation not supported
    ' 2+ = GDI+ used to display specific frame. No frame navigation supported
    ' When navigation is supported, the stdPicture.hPal property is repurposed.
    '   Set the current frame index with stdPicture.hPal, i.e., Me.Picture.hPal = 3
    '   Get the current frame index and total count with the same property:
    '       lValue = Me.Picture.hPal
    '       Count = lValue \ &H10000 (count is in the HiWord)
    '       Frame = lValue And &HFFFF& (index is in the LoWord)
    '   Frame wrapping applies in a simplified way. If you set a frame greater
    '   than the count, the first frame is used. If you set a frame lower than
    '   one, the last frame is used. This could be helpful in animation without
    '   needing to check for passing the min/max frames, for example:
    '       Me.Picture.hPal = (Me.Picture.hPal And &HFFFF&) + 1

    Dim tPic As StdPicture

    If ImageIndex = 1 Then
        If mgmtStyle <> msAutoSelect Or FillColor <> -1 Then
            ImageIndex = 0
        ElseIf alphaUsage <> acApplyAlphaChannel Or rndrAttrs <> 0 Then
            ImageIndex = 0
        ElseIf pvCanaryExists() = False Then
            ImageIndex = 0
        End If
    ElseIf ImageIndex = 0 Then
        If mgmtStyle = msAutoSelect And alphaUsage = acApplyAlphaChannel Then
            If FillColor = -1 And rndrAttrs = 0 Then
                If Width = 0 And Height = 0 Then Set tPic = pvStreamToPicture(oStream)
            End If
        End If
    End If
    If tPic Is Nothing Then
        Set tPic = pvLoadFormatGdiPlus(oStream, Width, Height, mgmtStyle, FillColor, alphaUsage, typeGIF, rndrAttrs, ImageIndex)
    End If
    Set pvLoadFormatGIF = tPic

End Function

Private Function pvLoadFormatIcon(arrData() As Byte, Width As Long, Height As Long, _
                            FillColor As Long, mgmtStyle As ManagedStyle, _
                            IconIndex As Long, IconDepth As Long, _
                            alphaUsage As AlphaUsageEnum, hAttrs As Long) As StdPicture
    
    ' Both VB & GDI+ have problems with icons, so we do it manually up to a point.
    ' This routine verifies passed array is icon/cursor format and optionally
    ' selects best image from a multi-image source.
    
    ' This routine always creates HICON/HCURSOR. If the goal is to have it
    ' managed or resized, then the icon/cursor handle is passed to another
    ' routine for that purpose.

    Dim hIcon As Long, lItemCount As Long
    Dim lIndex As Long, lValue As Long
    Dim lSize As Long, oStream As stdole.IUnknown
    Dim aSort() As Long, pProps() As Long, lOffsets() As Long
    Dim n As Long, x As Long, y As Long
    Dim oPic As StdPicture, tPic As StdPicture
    
    ' /// get best match and create picture
    lSize = UBound(arrData) + 1
    If lSize < 56& Then GoTo exitRoutine
    If Not (arrData(0) = 0 And arrData(1) = 0) Then GoTo exitRoutine    ' 1st integer must be zero
    If Not (arrData(2) = 1 Or arrData(2) = 2) Then GoTo exitRoutine     ' 2nd integer must be 0 or 1
    lItemCount = arrData(4) Or arrData(5) * &H100&
    lValue = lItemCount * 16& + 6&
    If lValue > lSize Or lValue = 6& Then GoTo exitRoutine
    ReDim pProps(0 To 2, 1 To lItemCount) ' width, height, bit depth
    
    ' continue with validation
    ReDim lOffsets(0 To 2, 0 To lItemCount - 1)
    n = 14&
    For lItemCount = 0& To lItemCount - 1&
        ' read each size/offset value & validate within range of file size
        CopyMemory lOffsets(0, lItemCount), arrData(n), 8&
        If lOffsets(0, lItemCount) < 0& Or lOffsets(1, lItemCount) < lValue Then GoTo exitRoutine ' invalid size/offset
        ' calc the start and end of each data block
        lOffsets(2, lItemCount) = lOffsets(0, lItemCount) + lOffsets(1, lItemCount)
        If lOffsets(2, lItemCount) > lSize Then GoTo exitRoutine
        n = n + 16&
    Next
    On Error GoTo 0
    
    ' simple loop to ensure no data block starts inside another; looking for corruption
    For y = 0& To lItemCount - 2&
        For x = y + 1& To lItemCount - 1&
            If (lOffsets(0, y) > lOffsets(0, x)) Then
                If (lOffsets(2, y) < lOffsets(0, x)) Then GoTo exitRoutine
            End If
        Next
    Next
    ' validation complete
    
    For lItemCount = 0& To lItemCount - 1&
        y = lOffsets(1, lItemCount)             ' offset to bitmapinfo header
        CopyMemory lValue, arrData(y), 4&       ' test for PNG
        If lValue = &H474E5089 Then             ' get dimensions/depth from PNG header
            CopyMemory x, arrData(y + 16&), 4&
            pProps(0, lItemCount + 1&) = pvReverseLong(x)
            CopyMemory x, arrData(y + 20&), 4&
            pProps(1, lItemCount + 1&) = pvReverseLong(x)
            n = arrData(lItemCount * 16& + 12&) ' use icondir bitcount if reasonable
            If (n = 0& Or n > 32&) Or arrData(2) = 2 Then ' else fetch it from header
                Select Case arrData(y + 25&)
                    Case 4, 6: n = 32&
                    Case 2: n = 24&
                    Case Else: n = arrData(y + 24&)
                End Select
            End If
            pProps(2, lItemCount + 1&) = n
        Else                                ' get real width/height
            CopyMemory pProps(0, lItemCount + 1&), arrData(y + 4&), 4&
            CopyMemory x, arrData(y + 8&), 4&
            pProps(1, lItemCount + 1&) = x \ 2&
            CopyMemory pProps(2, lItemCount + 1&), arrData(y + 14&), 2&
        End If
    Next
    
    ' sanity checks & set desired size
    If IconDepth < 0& Then IconDepth = 0&
    If Width = 0 Or Height = 0 Then
        If IconIndex = 0& Then              ' use system settings
            Width = GetSystemMetrics(11)    ' SM_CXICON
            Height = GetSystemMetrics(12)   ' SM_CYICON
        ElseIf IconIndex < 0& Then
            Width = GetSystemMetrics(49)    ' SM_CXSMICON
            Height = GetSystemMetrics(50)   ' SM_CYSMICON
        End If
    End If

    If lItemCount = 1& Then     ' no look-up needed
        lIndex = 0&
    ElseIf IconIndex > 0& Then  ' use passed index, no look-up needed
        If IconIndex > lItemCount Then lIndex = lItemCount - 1& Else lIndex = IconIndex - 1&
        
    Else                        ' find best match, simple sort algo
        ReDim aSort(0 To 1, 0 To lItemCount - 1&)
        For lItemCount = 0& To lItemCount - 1&
            lValue = (pProps(0, lItemCount + 1&) + pProps(1, lItemCount + 1&)) * &H100& Or pProps(2, lItemCount + 1&)
            For x = 0& To lItemCount - 1&       ' add to array, sorted
                If aSort(1, x) > lValue Then
                    For y = lItemCount To x + 1 Step -1&
                        aSort(0, y) = aSort(0, y - 1)
                        aSort(1, y) = aSort(1, y - 1)
                    Next
                    Exit For
                End If
            Next
            aSort(0, x) = lItemCount: aSort(1, x) = lValue
        Next
        ' note: when sorted, bits 0xFFFFFF00 contain size and 0xFF contains bit depth
        ' this produces the sorted list first by size then by depth, within duplicate sizes
        
        ' now find best size from the sorted list, exclude the depth for this part
        lValue = (Abs(Width) + Abs(Height)) * &H100&
        For x = 0& To lItemCount - 1&
            If (aSort(1, x) And &HFFFFFF00) >= lValue Then Exit For
        Next
        If x < lItemCount Then
            ' now find the best bit depth
            If IconDepth = 0& Then Call pvGetSysDPI(IconDepth)
            lValue = lValue Or IconDepth
            If (aSort(1, x) And &HFF) < IconDepth Then
                For x = x + 1& To lItemCount - 1&
                    If (aSort(1, x) And &HFFFFFF00) > (aSort(1, x - 1&) And &HFFFFFF00) Then
                        x = x - 1&: Exit For
                    End If
                    If (aSort(1, x) And &HFF) >= IconDepth Then Exit For
                Next
            End If
        End If
        If x = lItemCount Then lIndex = aSort(0, x - 1&) Else lIndex = aSort(0, x)
        Erase aSort()
    End If
    
    IconIndex = lItemCount                  ' return icon/cursor count
    
'    Debug.Print "Winner index:"; lIndex; " @ "; pProps(0, lIndex + 1); " x"; pProps(1, lIndex + 1); " bpp:"; pProps(2, lIndex + 1)
    
    If arrData(2) = 2 Then  ' cursor, hotspots need a bit of tweaking
        CopyMemory n, arrData(lOffsets(1, lIndex) - 4&), 4&
        CopyMemory arrData(lOffsets(1, lIndex) - 4&), arrData(lIndex * 16& + 10&), 4&
        hIcon = CreateIconFromResourceEx(arrData(lOffsets(1, lIndex) - 4&), lOffsets(0, lIndex) + 4&, 0&, &H30000, 0, 0, 0&)
        CopyMemory arrData(lOffsets(1, lIndex) - 4&), n, 4&
    Else
        hIcon = CreateIconFromResourceEx(arrData(lOffsets(1, lIndex)), lOffsets(0, lIndex), 1&, &H30000, 0, 0, 0&)
    End If
    If hIcon = 0 Then   ' XP/Win2K can't do PNG icons, apply workaround
        CopyMemory lValue, arrData(lOffsets(1, lIndex)), 4& ' bitmap header first 4 bytes
        If lValue = &H474E5089 Then  ' PNG
            Set oStream = pvStreamFromArray(arrData(), lOffsets(1, lIndex), lOffsets(0, lIndex))
            If oStream Is Nothing Then GoTo exitRoutine
            If GdipLoadImageFromStream(ObjPtr(oStream), lValue) <> 0 Then GoTo exitRoutine
            GdipCreateHICONFromBitmap lValue, hIcon
            GdipDisposeImage lValue: Set oStream = Nothing
        End If
    End If
    Erase lOffsets()
    If hIcon <> 0 Then
        Set oPic = pvHandleToPicture(hIcon, vbPicTypeIcon)
        If mgmtStyle = msManagedBitmap Then
            If pvCanaryExists() = False Then mgmtStyle = msUnmanagedBitmap
        End If
        If mgmtStyle = msAutoSelect And hAttrs = 0 And FillColor = -1 Then
            If Width = pProps(0, lIndex + 1) And Height = pProps(1, lIndex + 1) Then GoTo exitRoutine
        End If
        Set oPic = pvProcessHICON(oPic, Width, Height, FillColor, mgmtStyle, alphaUsage, hAttrs, False)
    End If
    Erase pProps()

exitRoutine:
    Set pvLoadFormatIcon = oPic

End Function

Private Function pvLoadFormatBitmap(oStream As stdole.IUnknown, _
                Width As Long, Height As Long, FillColor As Long, mgmtStyle As ManagedStyle, _
                alphaUsage As AlphaUsageEnum, rndrAttrs As Long) As StdPicture
    
    ' Function will use VB to load the bitmap only when alpha channel doesn't
    ' apply and when not resizing. Otherwise, GDI+ is used to load the picture
    
    Dim tPic As StdPicture

    If Width = 0 And Height = 0 Then
        If alphaUsage = acIgnoreAlphaChannel And rndrAttrs = 0 Then
            Set tPic = pvStreamToPicture(oStream)   ' fail over to GDI+ if needed
        End If
    End If
    If tPic Is Nothing Then ' use GDI+
        Set pvLoadFormatBitmap = pvLoadFormatGdiPlus(oStream, Width, Height, mgmtStyle, FillColor, alphaUsage, typeBitmap, rndrAttrs)
    Else
        Set pvLoadFormatBitmap = pvProcessBitmap(tPic, Width, Height, -1, mgmtStyle, alphaNone, rndrAttrs, True)
    End If

End Function

Private Function pvLoadFormatJPEG(oStream As stdole.IUnknown, Width As Long, _
                                Height As Long, FillColor As Long, _
                                mgmtStyle As ManagedStyle, alphaUsage As AlphaUsageEnum, _
                                rndrAttrs As Long) As StdPicture
                
    ' GDI+ used to load the JPG and falls back to VB if GDI+ fails.
                
    ' attempts to overcome two specific issues that VB can't handle
    ' 1) JPG in CMYK color space. VB fails to load these (invalid picture)
    ' 2) Non-standard orientation which is only available via the JPG's metadata which VB does not process
    
    Dim lSize As Long, lValue As Long
    Dim aData() As Byte, hDIB As Long
    Dim lOrient As Long, gHandle As Long
    Dim tPic As StdPicture, fRect(0 To 3) As Single
    
    Const PropertyTagOrientation  As Long = &H112&
    Const PixelFormatCMYK As Long = &H200F

    If GdipLoadImageFromStream(ObjPtr(oStream), gHandle) <> 0 Then
        IStream_Reset oStream
        Set tPic = pvStreamToPicture(oStream)
    Else
        GdipGetImageBounds gHandle, fRect(0), lValue
        If fRect(2) = 0 Or fRect(3) = 0 Then
            GdipDisposeImage gHandle: gHandle = 0
            IStream_Reset oStream
            Set tPic = pvStreamToPicture(oStream)
        End If
    End If
    If tPic Is Nothing And gHandle <> 0 Then
        GdipGetImagePixelFormat gHandle, lValue
        If lValue = PixelFormatCMYK Then ' deal with CMYK color space
            If GdipCreateHBITMAPFromBitmap(gHandle, hDIB, vbBlack) Then GoTo exitRoutine
        End If                          ' test for non-standard orientation
        GdipGetPropertyItemSize gHandle, PropertyTagOrientation, lSize
        If lSize = 18& Then
            ReDim aData(0 To lSize - 1&)
            If GdipGetPropertyItem(gHandle, PropertyTagOrientation, lSize, aData(0)) = 0& Then
                CopyMemory lSize, aData(4), 4&
                If lSize = 2& Then
                    CopyMemory lValue, aData(8), 4&
                    If lValue = 3& Then
                        lOrient = aData(16) Or aData(17) * &H100&
                        Select Case lOrient
                        Case 2, 4: lOrient = lOrient + 2    ' RotateNoneFlipX, Rotate180FlipX
                        Case 6, 8: lOrient = lOrient - 5    ' Rotate90FlipNone, Rotate270FlipNone
                        Case 1, 3: lOrient = lOrient - 1    ' Rotate180FlipNone
                        Case 5, 7                           ' Rotate90FlipX, Rotate270FlipX
                        Case Else: lOrient = 0              ' invalid value
                        End Select
                    End If
                End If
            End If
        End If
        If lOrient > 0 Then
            If hDIB <> 0 Then ' appears to be an issue flippping/rotating CMYK image, so...
                ' copy image to HBMP, then create new hImage from that
                GdipDisposeImage gHandle: gHandle = 0
                If GdipCreateBitmapFromHBITMAP(hDIB, 0&, gHandle) = 0& Then DeleteObject hDIB
            End If
            If gHandle <> 0 Then    ' alter the image & get new dimensions
                GdipImageRotateFlip gHandle, lOrient
                GdipGetImageBounds gHandle, fRect(0), lValue
            End If
        End If
    End If

exitRoutine:
    If tPic Is Nothing Then
        If alphaUsage = acApplyAlphaChannel Or alphaUsage = acIgnoreAlphaChannel Then
            hDIB = pvDIBFromGDIplus(gHandle, Width, Height, FillColor, , , fRect())
        Else
            hDIB = pvDIBFromGDIplus(gHandle, (fRect(2)), (fRect(3)), vbWhite, , , fRect())
        End If
        GdipDisposeImage gHandle
        If hDIB = 0 Then Exit Function
        Set tPic = pvHandleToPicture(hDIB, vbPicTypeBitmap)
    End If
    Set pvLoadFormatJPEG = pvProcessBitmapAlpha(tPic, Width, Height, FillColor, mgmtStyle, alphaUsage, rndrAttrs, True)
    
End Function

Private Function pvLoadFormatGdiPlus(oStream As stdole.IUnknown, _
                        Width As Long, Height As Long, mgmtStyle As ManagedStyle, _
                        ByVal FillColor As Long, alphaUsage As AlphaUsageEnum, _
                        lType As ePicType, rndrAttrs As Long, _
                        Optional FrameIndex As Long) As StdPicture
                        
    ' Routine loads a GDI+ object from stream and converts to bitmap
    ' Depending on parameters, picture is handled within this routine
    '   or passed to pvProcessBitmapAlpha
                        
    ' FrameIndex: applies only to GIF/TIFF
    '   TIFF: page to return, on exit FrameIndex=NrPages
    '   GIF: 0 = 1st frame, no frame navigation: on exit FrameIndex=NrFrames
    '        1 = 1st frame, frame navigation: on exit FrameIndex = FrameIndex(loWord) & FrameCount(hiWord)
    '            if only 1 frame exists, treated as if FrameIndex=0
    '       >1 = frame to return, no frame navigation: on exit FrameIndex=NrFrames

    Dim gHandle As Long, hDIB As Long
    Dim lColorType As Long, aSource() As Long
    Dim fRect(0 To 4) As Single ' fRect(4) = generic Single variable
    Dim hDC As Long, oPic As StdPicture
    Dim lValue As Long, gHookHandle As Long
    Const PixelFormat16bppARGB As Long = &H61007
    
    If GdipLoadImageFromStream(ObjPtr(oStream), gHandle) <> 0 Then Exit Function
    GdipGetImageBounds gHandle, fRect(0), lValue
    If fRect(2) < 1 Or fRect(3) < 1 Then Exit Function
    
    If lType = typeMetafile Then ' scale metafile dimensions as needed
        ' only get here in 1 case: initially loaded by user & not wanting metafile format
        If Width = 0 Then
            GdipGetImageHorizontalResolution gHandle, fRect(4)
            If fRect(4) = 0 Then fRect(4) = 96
            Width = fRect(2) * 96 / fRect(4)
        End If
        If Height = 0 Then
            GdipGetImageVerticalResolution gHandle, fRect(4)
            If fRect(4) = 0 Then fRect(4) = 96
            Height = fRect(3) * 96 / fRect(4)
        End If
        ' convert metafile to 32bpp DIB with valid alpha channel
        pvCreateDibDC hDIB, Abs(Width), Abs(Height), 32, hDC
        If hDIB = 0 Then
            GdipDisposeImage gHandle: GoTo exitRoutine
        End If
        lValue = pvSetHGraphics(hDC, hDIB, vbBlack, 0, True)
        If lValue = 0 Then
            DeleteObject hDIB: DeleteDC hDC
            GdipDisposeImage gHandle: GoTo exitRoutine
        End If
        GdipDrawImageRectRectI lValue, gHandle, 0, 0, Abs(Width), Abs(Height), fRect(0), fRect(1), fRect(2), fRect(3), UnitPixels, 0, 0, 0
        GdipDeleteGraphics lValue: GdipDisposeImage gHandle
        hDIB = SelectObject(hDC, hDIB): DeleteDC hDC
        Set oPic = pvConvertMetafile(oPic, Width, Height, FillColor, mgmtStyle, alphaUsage, rndrAttrs, hDIB)
        GoTo exitRoutine
    End If
    
    If Width = 0 Then Width = fRect(2)
    If Height = 0 Then Height = fRect(3)
    
    If lType = typeGIF Then ' get frame count & setup frame navigation if applies
        lValue = pvGetFramePageCount(gHandle, lType, FrameIndex)
        If mgmtStyle = msAutoSelect And FrameIndex = 1 And lValue > 1 Then
            ' allow frame navigation
            FrameIndex = lValue ' return frame count
            lValue = lValue * &H10000 Or &H80000001
            ' thunk uses rendering flags for sizing/mirroring - set rendering flags
            If Width < 0 Then lValue = lValue Or &H2000
            If Abs(Width) <> fRect(2) Then lValue = lValue Or &H1000
            If Height < 0 Then lValue = lValue Or &H4000
            If Abs(Height) <> fRect(3) Then lValue = lValue Or &H1000
            ' create VB picture placeholder
            hDIB = pvDIBFromGDIplus(gHandle, Abs(Width), Abs(Height), vbWhite, gHookHandle, , fRect(), PixelFormat24bppRGB)
            If hDIB = 0 Then    ' failure, abort
                GdipDisposeImage gHandle: gHandle = 0
            Else                ' hook the picture
                If gHookHandle <> 0 Then GdipDisposeImage gHookHandle
                Set oPic = pvHandleToPicture(hDIB, vbPicTypeBitmap)
                pvHookPicture oPic, gHandle, typeGIF, lValue: gHandle = 0
            End If
            GoTo exitRoutine
        End If
        FrameIndex = lValue ' return frame count
        
    ElseIf lType = typeTIFF Then ' get page count & set initial page
        FrameIndex = pvGetFramePageCount(gHandle, typeTIFF, FrameIndex)
        
    ElseIf lType = typeUnknown Then
        GdipGetImageType gHandle, lValue
        If lValue <> 1 Then ' sanity check. Whatever this is, make a GDI+ bitmap
            GdipCreateBitmapFromScan0 fRect(2), fRect(3), 0, PixelFormat32bppARGB, ByVal 0, gHookHandle
            If gHookHandle = 0 Then GoTo processBitmap
            lValue = pvSetHGraphics(0&, gHookHandle, -1, 0, False, False)
            GdipDrawImageRectRectI lValue, gHandle, 0, 0, fRect(2), fRect(3), fRect(0), fRect(1), fRect(2), fRect(3), UnitPixels, 0, 0, 0
            GdipDeleteGraphics lValue
            GdipDisposeImage gHandle: gHandle = gHookHandle: gHookHandle = 0
            fRect(0) = 0: fRect(1) = 0
        End If
    End If
    
    GdipGetImagePixelFormat gHandle, lColorType
    lValue = (lColorType And &HFF00&) \ &H100
    If lValue > 32 Then ' should future versions of GDI+ support 48/64 bpp images
        lColorType = PixelFormat32bppARGB
    End If
    If lType = typeBitmap Then  ' test for GDI+ loading of v5 bitmap header & non-alpha bmps
        If lColorType >= alphaExists Then
            lType = typeUnknown ' don't process as normal bmp
        ElseIf lValue < 32 Then ' non-alpha compatible, change alpha usage if appropriate
            If alphaUsage = acApplyAlphaChannel Or alphaUsage = acWantFormat32bppPARGB Then
                alphaUsage = acIgnoreAlphaChannel
            End If
        End If
    End If
    
    If (lType = typeBitmap And alphaUsage <> acIgnoreAlphaChannel) Or _
       (lType <> typeBitmap And ((alphaUsage And Not acWantFormat32bppPARGB) <> acApplyAlphaChannel)) Then
        ' bitmap with unknown possible alpha or a format ignoring alpha or using color key
        ' in either case, get bits to RGB & let pvProcessBitmapAlpha handle it
        lValue = pvCreateDibDC(hDIB, fRect(2), fRect(3), 32)
        If hDIB <> 0 Then pvConvertFormat gHandle, PixelFormat32bppRGB, lValue, fRect(2) * -4
    
    Else ' all other cases, we finish off & done
         ' alpha being applied or converted to PARGB or alpha not applicable
        If (alphaUsage And acWantFormat32bppPARGB) = 0 Then
            If lColorType = PixelFormat16bppARGB Then lColorType = PixelFormat32bppARGB
        Else
            lColorType = PixelFormat32bppPARGB
        End If
        hDIB = pvDIBFromGDIplus(gHandle, Width, Height, FillColor, gHookHandle, rndrAttrs, fRect(), lColorType)
        GdipDisposeImage gHandle: gHandle = 0
        If hDIB = 0 Then GoTo exitRoutine
        Set oPic = pvHandleToPicture(hDIB, vbPicTypeBitmap)
        If mgmtStyle = msAutoSelect Then
            GdipGetImagePixelFormat gHookHandle, lColorType
            If lColorType >= alphaExists Or rndrAttrs <> 0 Then
                ReDim aSource(0 To 5)
                GetGDIObject hDIB, 24, aSource(0)
                If pvValidateAlphaChannel(aSource(5), Width * Height) <> alphaNone Then
                    mgmtStyle = msManagedBitmap
                End If
                Erase aSource()
            End If
        End If
        If mgmtStyle = msManagedBitmap Then
            If pvCanaryExists() = True Then
                pvHookPicture oPic, gHookHandle: gHookHandle = 0
            End If
        End If
        If gHookHandle <> 0 Then GdipDisposeImage gHookHandle
        GoTo exitRoutine
    End If
    
processBitmap:
    If gHandle <> 0 Then GdipDisposeImage gHandle: gHandle = 0
    If hDIB <> 0 Then
        Set oPic = pvHandleToPicture(hDIB, vbPicTypeBitmap)
        Set oPic = pvProcessBitmapAlpha(oPic, Width, Height, FillColor, mgmtStyle, alphaUsage, rndrAttrs, True)
    End If
exitRoutine:
    Set pvLoadFormatGdiPlus = oPic

End Function

Private Function pvConvertFormat(gSource As Long, newFormat As Long, _
                        dstPointer As Long, lStride As Long) As Boolean

    ' converts GDI+ image pixel format to another and writes to passed buffer
    ' only used for conversion between the three 32bpp formats
    Dim uBmp As BitmapData, fRect(0 To 3) As Single, lValue As Long
    Const ReadToBuffer As Long = &H5

    GdipGetImageBounds gSource, fRect(0), lValue
    With uBmp
        .bpHeight = fRect(3): .bpWidth = fRect(2)
        .bpStride = lStride: .bpPixelFormat = newFormat
        If lStride < 0 Then
            .bpScan0Ptr = ((dstPointer Xor &H80000000) + lStride * -(.bpHeight - 1)) Xor &H80000000
        Else
            .bpScan0Ptr = dstPointer
        End If
        GdipBitmapLockBits gSource, .bpLeft, ReadToBuffer, .bpPixelFormat, .bpWidth
        GdipBitmapUnlockBits gSource, .bpWidth
    End With

End Function

Private Function pvConvertGIF(tPic As Picture, Width As Long, Height As Long, _
                             FillColor As Long, mgmtStyle As ManagedStyle, _
                             alphaUsage As AlphaUsageEnum, hAttrs As Long) As StdPicture
                            
    ' used to convert an existing stdPicture GIF to a 32bppPARGB bitmap for
    ' scaling and to support transparency. GIFs are tricky. In the IDE, we can
    ' get their original format from the stdPicture and parse that since it's
    ' a true GIF format. However, when compiled, that original format doesn't
    ' exist any longer. Therefore, we need to find out the transparency color
    ' index used in the picture's palette...
                            
    Dim oPic As StdPicture, aSource(0 To 5) As Long
    Dim aData() As Long, aBits() As Long
    Dim hDIB As Long, lValue As Long, lAlpha As eAlphaType
    Dim gHandle As Long, lFlags As Long
    Dim lColor As Long, hDC As Long, x As Long

    Me.GetImageDimensions tPic, aSource(1), aSource(2)
    If Width = 0 Then Width = aSource(1)
    If Height = 0 Then Height = aSource(2)

    ' find transparent color if any, by using a color not in the palette
    hDC = GetDC(0)
    ReDim aData(0 To 296): aData(0) = 40
    GetDIBits hDC, tPic.Handle, 0, 0, ByVal 0&, aData(0), 0
    lValue = (aData(3) And &HFF0000) \ &H10000: aData(4) = 0
    If lValue <= 8 Then               ' get palette entries
        aData(4) = 0: aData(5) = 0
        GetDIBits hDC, tPic.Handle, 0, 0, ByVal 0&, aData(0), 0
    End If
    ReleaseDC 0, hDC
    If lValue > 8 Then Exit Function ' should never happen
    lColor = 1908: lValue = 2 ^ lValue ' nr of palette entries
    Do
        For x = 0 To lValue - 1
            If aData(10 + x) = lColor Then Exit For
        Next
        If x = lValue Then Exit Do
        lColor = lColor + 1908    ' keep looping until color not in palette
    Loop
    Erase aData()

    hDIB = pvDIBFromStdPicture(tPic, aSource(1), aSource(2), lColor, aSource(5))
    If hDIB = 0 Then Exit Function
    ' test color is BGR, convert to RGB & find color
    pvSetOverlay aData(), VarPtrArray(aBits), aSource(5), aSource(1) * aSource(2)
    lColor = (lColor And &HFF00&) Or (lColor And &HFF) * &H10000 Or (lColor And &HFF0000) \ &H10000
    For lValue = 0 To UBound(aBits)
        If aBits(lValue) = lColor Then
            aBits(lValue) = 0: lAlpha = PixelFormat32bppARGB
        Else
            aBits(lValue) = aBits(lValue) Or &HFF000000
        End If
    Next
    pvSetOverlay aData(), VarPtrArray(aBits), 0, 0
    
setReturn:
    Set oPic = pvHandleToPicture(hDIB, vbPicTypeBitmap)
    Set pvConvertGIF = pvProcessBitmapAlpha(oPic, Width, Height, FillColor, mgmtStyle, alphaUsage, hAttrs, True)
                            
End Function

Private Function pvConvertMetafile(tPic As Picture, Width As Long, Height As Long, _
                                FillColor As Long, mgmtStyle As ManagedStyle, _
                                alphaUsage As AlphaUsageEnum, hAttrs As Long, _
                                Optional gdipDIB As Long) As StdPicture
                            
    ' used to convert WMF/EMF to bitmap for scaling and/or filling
                            
    Dim oPic As StdPicture, aSource(0 To 5) As Long
    Dim aData() As Long, aBits() As Long
    Dim hDIB As Long, lValue As Long
    
    If gdipDIB = 0 Then
        Me.GetImageDimensions tPic, aSource(1), aSource(2)
        If Width = 0 Then Width = aSource(1)
        If Height = 0 Then Height = aSource(2)
        ' convert metafile w/transparency to 32bpp DIB
        hDIB = pvDIBFromStdPicture(tPic, Abs(Width), Abs(Height), vbBlack, aSource(5))
        If hDIB = 0 Then Exit Function
    
    Else ' passed from pvLoadFormatGdiPlus
        hDIB = gdipDIB
        GetGDIObject hDIB, 24, aSource(0)
    End If
    
    pvSetOverlay aData(), VarPtrArray(aBits), aSource(5), Abs(Width * Height)
    ' when metafile renders, it toggles alpha of pixels it touched
    For lValue = 0 To UBound(aBits)
        aBits(lValue) = aBits(lValue) Xor &HFF000000
    Next
    pvSetOverlay aData(), VarPtrArray(aBits), 0, 0
    
exitRoutine:
    Set oPic = pvHandleToPicture(hDIB, vbPicTypeBitmap)
    Set pvConvertMetafile = pvProcessBitmapAlpha(oPic, Width, Height, FillColor, mgmtStyle, alphaUsage, hAttrs, False)
                            
End Function

Private Function pvProcessBitmap(tPic As StdPicture, Width As Long, Height As Long, _
                    FillColor As Long, mgmtStyle As ManagedStyle, _
                    AlphaType As eAlphaType, rndrAttrs As Long, _
                    Optional forceNewBmp As Boolean = True) As StdPicture

    ' Core routine for manipulating non-icon/cursor. pvProcessHICON used for icons/cursors
    ' Most params passed from CopyPicture,LoadPictureEx,LoadResPictureEx,ManagePicture
    ' AlphaType: if non-zero, passed from pvProcessBitmapAlpha or pvConvertxxx routines
    ' forceNewBmp: If passed as true or set to true below, then returned picture
    '   will not be the same one passed to this routine. CopyPicture sets this to true

    Dim lValue As Long, oPic As StdPicture
    Dim gHandle As Long, mHandle As Long
    Dim aSource(0 To 5) As Long, aData(0 To 9) As Long
    Dim aBits() As Long, hDIB As Long, gHookHandle As Long
    
    '/// Step 1: Get picture properties & actions needed
    GetGDIObject tPic.Handle, 24, aSource(0)
    aSource(4) = (aSource(4) And &HFF0000) \ &H10000 ' bit depth
    pvIsManaged tPic, , mHandle  ' currently managed?
    If mHandle <> 0 Then GdipGetImagePixelFormat mHandle, AlphaType
    If Width = 0 Then Width = aSource(1)
    If Height = 0 Then Height = aSource(2)
    If mgmtStyle = msAutoSelect Then
        If FillColor <> -1 Then
            mgmtStyle = msUnmanagedBitmap   ' end result = opaque
        ElseIf rndrAttrs = 0 And AlphaType < alphaExists Then
            mgmtStyle = msUnmanagedBitmap   ' end result = opaque
        Else
            mgmtStyle = msManagedBitmap     ' may have transparency
        End If
    End If
    If mgmtStyle = msManagedBitmap Then     ' thunk required
        If pvCanaryExists() = False Then mgmtStyle = msUnmanagedBitmap
    End If
    If aSource(4) = 32 Then
        If AlphaType = alphaNone Then AlphaType = PixelFormat32bppRGB
    ElseIf aSource(4) = 24 Then
        If AlphaType = alphaNone Then AlphaType = PixelFormat24bppRGB
    Else ' 24 or 32bpp for management, AlphaType filled in later
        If mgmtStyle = msManagedBitmap Then forceNewBmp = True
    End If
    If forceNewBmp = False Then             ' need new bmp to render to?
        If rndrAttrs <> 0 Then
            forceNewBmp = True              ' attrs, render required
        ElseIf Width <> aSource(1) Or Height <> aSource(2) Then
            forceNewBmp = True              ' resizing/mirroring
        ElseIf AlphaType >= alphaExists And FillColor <> -1 Then
            forceNewBmp = True              ' filling, render required
        End If
    End If
    If forceNewBmp = False Then
        Set oPic = tPic: GoTo setReturn
    End If
    
    '/// Step 2: Create GDI+ around bitmap for rendering
    If mHandle = 0 Then
        If aSource(5) <> 0 And (aSource(4) = 24 Or aSource(4) = 32) Then
            ' wrap GDI+ around DIB bits
            aSource(5) = ((aSource(5) Xor &H80000000) + aSource(3) * (aSource(2) - 1)) Xor &H80000000
            GdipCreateBitmapFromScan0 aSource(1), aSource(2), -aSource(3), AlphaType, ByVal aSource(5), gHandle
        Else                                    ' get bits into array, wrap GDI+ on array
            ReDim aBits(0 To aSource(1) * aSource(2) - 1)
            aData(0) = 40: aData(1) = aSource(1): aData(2) = -aSource(2): aData(3) = &H200001
            lValue = GetDC(0)
            GetDIBits lValue, tPic.Handle, 0, aSource(2), aBits(0), aData(0), 0
            ReleaseDC 0, lValue: Erase aData()
            aSource(5) = VarPtr(aBits(0))
            aSource(3) = aSource(1) * 4: aSource(4) = 32
            If AlphaType = PixelFormat24bppRGB Then AlphaType = PixelFormat32bppRGB
            GdipCreateBitmapFromScan0 aSource(1), aSource(2), aSource(3), AlphaType, ByVal aSource(5), gHandle
        End If
        If gHandle = 0 Then GoTo exitRoutine
    Else
        gHandle = mHandle                       ' use existing manged GDI+ handle
    End If
    
    '/// Step 3: render into new DIB, sizing, mirroring, filling as needed
    hDIB = pvDIBFromGDIplus(gHandle, Width, Height, FillColor, gHookHandle, rndrAttrs, , AlphaType)
    If mHandle = 0 Then GdipDisposeImage gHandle
    Erase aBits(): gHandle = 0
    If hDIB = 0 Then GoTo exitRoutine
    If mgmtStyle = msUnmanagedBitmap Then GdipDisposeImage gHookHandle: gHookHandle = 0
    
    '/// Step 4: set return picture and optionally manage it
    Set oPic = pvHandleToPicture(hDIB, vbPicTypeBitmap): hDIB = 0
setReturn:
    If mgmtStyle = msManagedBitmap And gHookHandle = 0 Then
        If gHandle <> 0 Then GdipDisposeImage gHandle: gHandle = 0
        lValue = ((aSource(5) Xor &H80000000) + aSource(3) * (Height - 1)) Xor &H80000000
        GdipCreateBitmapFromScan0 Width, Height, -aSource(3), AlphaType, ByVal lValue, gHookHandle
    End If
    If gHookHandle <> 0 Then pvHookPicture oPic, gHookHandle

exitRoutine:                                    ' clean up
    If oPic Is Nothing Then Set oPic = tPic
    If gHandle <> 0 Then GdipDisposeImage gHandle
    If hDIB <> 0 Then DeleteObject hDIB
    Set pvProcessBitmap = oPic

End Function

Private Function pvProcessBitmapAlpha(tPic As StdPicture, Width As Long, Height As Long, _
                    FillColor As Long, mgmtStyle As ManagedStyle, alphaUsage As AlphaUsageEnum, _
                    ByVal rndrAttrs As Long, _
                    Optional CanOverwrite As Boolean = False) As StdPicture

    ' This routine is called for each newly loaded bitmap, when requested
    ' Also called when requested via ManagePicture
    
    ' Routine creates/tests 32bpp image if alpha transparency applies. Else aborts
    ' All parameters are passed from LoadPictureEx, LoadResPictureEx or ManagePicture
    ' CanOverwrite is true if passed tPic can be directly written to, otherwise
    '   a new image is returned
    
    Dim oPic As StdPicture
    If alphaUsage = acIgnoreAlphaChannel Then GoTo exitRoutine
    
    Dim lAlphaType As eAlphaType, cKey As Long
    Dim lValue As Long, aSource(0 To 5) As Long
    Dim aBits() As Long, aData() As Long, gHandle As Long
    Dim x As Long, hDIB As Long, pBits As Long
    Const ColorKeyMask = &H86FFFFFF

    ' /// Step 1: Gather image info & exit early if only applying alpha channel
    GetGDIObject tPic.Handle, 24, aSource(0)
    ' aSource(0)    zero
    '        (1)    width
    '        (2)    height
    '        (3)    scan width
    '        (4)    LoWord = planes, HiWord = bit depth
    '        (5)    DIB bits pointer, N/A for DDB
    aSource(4) = (aSource(4) And &HFF0000) \ &H10000
    If aSource(4) < 32 Then
        aSource(5) = 0          ' flag = copy bits if needed
    ElseIf pvIsManaged(tPic, , gHandle) <> 0 Then
        GdipGetImagePixelFormat gHandle, lAlphaType
        If alphaUsage = acApplyAlphaChannel Then GoTo exitRoutine
        If alphaUsage = acWantFormat32bppPARGB Then
            If lAlphaType = PixelFormat32bppPARGB Then GoTo exitRoutine
        End If
        aSource(5) = 0          ' flag = copy bits if needed
    End If
    If alphaUsage = acApplyAlphaChannel Then
        If aSource(4) = 32 And aSource(5) <> 0 Then ' 32 bit DIB vs DDB/managed
            lAlphaType = pvValidateAlphaChannel(aSource(5), aSource(1) * aSource(2))
            GoTo exitRoutine
        End If
    End If
    
    ' /// Step 2: Get bits into buffer, process the buffer
    ReDim aBits(0 To aSource(1) * aSource(2) - 1)
    aSource(0) = VarPtr(aBits(0))
    If gHandle = 0 Then
        lValue = GetDC(0)
        If lValue = 0 Then GoTo exitRoutine
        ReDim aData(0 To 9)
        aData(0) = 40: aData(1) = aSource(1): aData(2) = aSource(2): aData(3) = &H200001
        GetDIBits lValue, tPic.Handle, 0, aSource(2), aBits(0), aData(0), 0
        ReleaseDC 0, lValue: Erase aData()
    Else
        If lAlphaType < alphaExists Then lAlphaType = PixelFormat32bppRGB
        pvConvertFormat gHandle, lAlphaType, aSource(0), aSource(1) * 4
    End If
    
    ' /// Step 3: test alpha channel for 32bpp sources if not done above
    If (alphaUsage And acIgnoreAlphaChannel) = 0 Then
        If aSource(4) = 32 And lAlphaType = alphaNone Then
            lAlphaType = pvValidateAlphaChannel(aSource(0), aSource(1) * aSource(2))
            If alphaUsage = acApplyAlphaChannel Then
                If lAlphaType = alphaNone Then GoTo exitRoutine Else GoTo setReturn
            ElseIf alphaUsage = acWantFormat32bppPARGB Then
                If (lAlphaType And PixelFormat32bppPARGB) = PixelFormat32bppPARGB Then GoTo setReturn
            End If
        End If
    End If
    
    ' /// Step 4: convert to PARGB as needed
    If lAlphaType = PixelFormat32bppARGB Then
        If (alphaUsage And acWantFormat32bppPARGB) = acWantFormat32bppPARGB Then
            aData() = aBits()
            If GdipCreateBitmapFromScan0(aSource(1), aSource(2), aSource(1) * 4, PixelFormat32bppARGB, aData(0), lValue) = 0 Then
                pvConvertFormat lValue, PixelFormat32bppPARGB, aSource(0), aSource(1) * 4
                GdipDisposeImage lValue
                lAlphaType = PixelFormat32bppPARGB
            End If
            Erase aData()
            If alphaUsage = acWantFormat32bppPARGB Then GoTo setReturn
        End If
    End If
    
    ' /// Step 5: set the color key as needed
    If (alphaUsage And acColorKeyIsTopLeftPixel) = acColorKeyIsTopLeftPixel Then
        cKey = aBits(aSource(1) * (aSource(2) - 1)) And &HFFFFFF
    ElseIf (alphaUsage And acColorKeyIsBlack) = 0 Then ' else color key is black
        cKey = alphaUsage And &H80FFFFFF
        If cKey < 0 Then cKey = GetSysColor(cKey And &HFF)
    End If
    
    ' /// Step 6: process applicable alpha type: pARGB, ARGB, none/RGB
    If lAlphaType < alphaExists Then
        ' result will be simple transparency, similar to GIF transparency
        ' possible no color key hits - test for that
        lAlphaType = alphaNone
        For x = 0 To UBound(aBits)
            ' don't include alpha channel in color key (alpha may be excluded)
            If (aBits(x) And &HFFFFFF) = cKey Then
                aBits(x) = 0: lAlphaType = PixelFormat32bppUARGB ' flag = has alpha
            Else
                aBits(x) = aBits(x) Or &HFF000000   ' set alpha byte
            End If
        Next
    Else    ' color key in use; else handled earlier
        cKey = cKey Or &HFF000000  ' include alpha channel for searching
        For x = 0 To UBound(aBits)
            If aBits(x) = cKey Then aBits(x) = 0
        Next
    End If
    
setReturn:
    ' /// Final step: set return object
    If lAlphaType >= alphaExists Then
        If CanOverwrite = False Or aSource(5) = 0 Then
            aSource(5) = pvCreateDibDC(hDIB, aSource(1), aSource(2), 32)
            If hDIB = 0 Then
                lAlphaType = alphaNone: Erase aBits()
                GoTo exitRoutine
            End If
            CanOverwrite = True
        End If
        CopyMemory ByVal aSource(5), ByVal aSource(0), aSource(1) * aSource(2) * 4&
        If hDIB <> 0 Then Set oPic = pvHandleToPicture(hDIB, vbPicTypeBitmap)
    End If

exitRoutine:
    Erase aBits()
    If lAlphaType = PixelFormat32bppUARGB Then
        If (alphaUsage And acWantFormat32bppPARGB) = 0 Then
            lAlphaType = PixelFormat32bppARGB
        Else
            lAlphaType = PixelFormat32bppPARGB
        End If
    End If
    If oPic Is Nothing Then Set oPic = tPic
    Set pvProcessBitmapAlpha = pvProcessBitmap(oPic, Width, Height, FillColor, mgmtStyle, lAlphaType, rndrAttrs, Not CanOverwrite)

End Function

Private Function pvProcessHICON(tPic As Picture, Width As Long, Height As Long, _
                            FillColor As Long, mgmtStyle As ManagedStyle, _
                            alphaUsage As AlphaUsageEnum, hAttrs As Long, _
                            Optional forceNewIcon As Boolean = True, _
                            Optional hIcon As Long) As StdPicture

    ' This routine will convert HICON/HCURSOR handle to a 32bpp bitmap for
    ' scaling with GDI+ and if unmanaged will convert that bitmap back to
    ' a HICON/HCURSOR handle afterwards
    
    Dim aData() As Long, ICI(0 To 4) As Long
    Dim gHandle As Long, xRatio As Single
    Dim lValue As Long, gHookHandle As Long
    Dim hDIB As Long, x As Long, y As Long
    Dim aSource(0 To 5) As Long, uSafeArray() As Long
    Dim aBits() As Long, aMask() As Byte
    Dim oPic As StdPicture, hDC As Long
    Dim mPtr As Long, mValue As Long
    Dim mskShift As Long, mScanWidth As Long
    
    Me.GetImageDimensions tPic, aSource(1), aSource(2)
    aSource(3) = aSource(1) * 4
    If Width = 0 Then Width = aSource(1)
    If Height = 0 Then Height = aSource(2)
    If mgmtStyle = msManagedBitmap Then
        If pvCanaryExists() = False Then mgmtStyle = msUnmanagedBitmap
    End If
    If forceNewIcon = False Then
        If mgmtStyle <> msAutoSelect Then
            forceNewIcon = True
        ElseIf Height <> aSource(2) Or Width <> aSource(1) Then
            forceNewIcon = True
        ElseIf hAttrs <> 0 Or FillColor <> -1 Then
            forceNewIcon = True
        End If
    End If
    If forceNewIcon = False Then GoTo exitRoutine
    
    If tPic.Type = vbPicTypeBitmap Then
        ' passed from CopyPictureToHandle, wanting HICON from HBITMAP
        ICI(0) = 1: ICI(4) = tPic.Handle
    Else
        GetIconInfo tPic.Handle, ICI(0)
        ' ICI(0)    0=cursor,1=icon
        '    (1)    X hotspot
        '    (2)    Y hotspot
        '    (3)    Mask bitmap
        '    (4)    Color bitmap - zero if B/W icon/cursor
        If ICI(0) = 0 Then ' cursor
            If ICI(4) = 0 Then ' B&W cursor
                If mgmtStyle = msAutoSelect Then
                    hIcon = CopyImage(tPic.Handle, 2, Abs(Width), Abs(Height), 0)
                    Set oPic = pvHandleToPicture(hIcon, vbPicTypeIcon)
                    GoTo exitRoutine
                End If
            Else    ' mirror hot spot if applies
                If Width < 0 Then ICI(0) = 2
                If Height < 0 Then ICI(0) = ICI(0) Or 4
            End If
        End If
    End If
    
    ' build a 32bppARGB icon/bitmap in an array
    ReDim aBits(0 To aSource(1) * aSource(2) - 1)
    ReDim aData(0 To 9)
    aData(0) = 40: aData(1) = aSource(1): aData(2) = -aSource(2): aData(3) = &H200001
    hDC = GetDC(0)
    If ICI(4) = 0 Then                  ' B&W icon/cursor
        aData(2) = aData(2) + aData(2)
        GetDIBits hDC, ICI(4), 0, aSource(2), aBits(0), aData(0), 0
    Else                                ' validate alpha channel used
        GetDIBits hDC, ICI(4), 0, aSource(2), aBits(0), aData(0), 0
        If tPic.Type = vbPicTypeBitmap Then
            ' converting bmp to ico, ensure bits are ARGB
            x = pvValidateAlphaChannel(VarPtr(aBits(0)), aSource(1) * aSource(2))
            If x = PixelFormat32bppPARGB Then ' convert to ARGB
                aData() = aBits()
                GdipCreateBitmapFromScan0 aSource(1), aSource(2), aSource(3), x, aData(0), gHandle
                pvConvertFormat gHandle, PixelFormat32bppARGB, VarPtr(aBits(0)), aSource(3)
                GdipDisposeImage gHandle: gHandle = 0
                Erase aData()
            ElseIf x = alphaNone Then
                For x = 0 To UBound(aBits)
                    aBits(x) = aBits(x) Or &HFF000000
                Next
            End If
        Else
            For x = UBound(aBits) To 0 Step -1
                If (aBits(x) And &HFF000000) <> 0 Then Exit For
            Next
        End If
    End If
    If (x = -1 Or ICI(4) = 0) Then  ' get alpha from mask
        mScanWidth = ((aSource(1) + &H1F&) And Not &H1F&) \ &H8& ' mask scanwdith
        ReDim aMask(0 To mScanWidth * aSource(2)) ' purposely size 1 byte larger
        ReDim aData(0 To 11)    ' bitmap header + 2 colors
        aData(0) = 40: aData(1) = aSource(1): aData(2) = aSource(2)
        aData(3) = &H10001: aData(8) = 2: aData(9) = 2: aData(11) = vbWhite
        If ICI(4) = 0 Then      ' set offset within b/w icon/cursor
            y = aSource(2): aData(2) = aData(2) + aData(2)
        Else
            y = 0               ' non b/w icon/cursor offset
        End If                  ' get the mask bits (8 pixels per byte)
        GetDIBits hDC, ICI(3), y, aSource(2), aMask(0), aData(0), 0
        Erase aData(): lValue = 0
        For y = 0 To aSource(2) - 1
            ' set scanline related variables
            mPtr = (aSource(2) - y - 1) * mScanWidth
            mskShift = &H80: mValue = aMask(mPtr)
            x = y * aSource(1)  ' color pixel offset
            For x = x To x + aSource(1) - 1
                If (mValue And mskShift) = 0 Then  ' pixel opaque
                    aBits(x) = aBits(x) Or &HFF000000
                ElseIf aBits(x) = 0 Then
                    aBits(x) = 0: lValue = 1    ' pixel transprent
                Else                            ' else Xor with backcolor
                    aBits(x) = aBits(x) Xor FillColor Or &HFF000000
                End If
                If mskShift = 1 Then               ' 8 pixels done, next ptr
                    mskShift = &H80: mPtr = mPtr + 1
                    mValue = aMask(mPtr)        ' get next 8 mask bits
                Else
                    mskShift = mskShift \ 2
                End If
            Next
        Next
        Erase aMask()
    End If
    ReleaseDC 0, hDC
    If ICI(4) <> 0 Then
        If tPic.Type = vbPicTypeIcon Then DeleteObject ICI(4)
        ICI(4) = 0
    End If
    If ICI(3) <> 0 Then ' can we reuse the mask
        If Width <> aSource(1) Or Height <> aSource(2) Then
            lValue = 1
        Else
            If hAttrs <> 0 Or FillColor <> -1 Then lValue = 1 Else lValue = 0
        End If
        If lValue = 1 Or mgmtStyle <> msAutoSelect Then
            DeleteObject ICI(3): ICI(3) = 0
        End If
    End If
    
    ' create GDI+ overlay on array & DIB to render to
    GdipCreateBitmapFromScan0 aSource(1), aSource(2), aSource(3), PixelFormat32bppARGB, aBits(0), gHandle
    If gHandle = 0 Then
        Erase aBits(): GoTo exitRoutine
    End If
    If alphaUsage = acApplyAlphaChannel Or mgmtStyle = msAutoSelect Then
        ' alphaUsage ignored when returning icons, else applies if returning bitmap
        hDIB = pvDIBFromGDIplus(gHandle, Width, Height, FillColor, gHookHandle, hAttrs, , PixelFormat32bppARGB)
        GdipDisposeImage gHandle: gHandle = 0: Erase aBits()
        If mgmtStyle <> msAutoSelect Then       ' wanting bitmap
            Set oPic = pvHandleToPicture(hDIB, vbPicTypeBitmap)
            If mgmtStyle = msManagedBitmap Then
                pvHookPicture oPic, gHookHandle
            Else
                GdipDisposeImage gHookHandle
            End If
            hDIB = 0: GoTo exitRoutine
        End If
        GdipDisposeImage gHookHandle
    
    Else ' bitmap where alphaUsage <> acApplyAlphaChannel
        ' create DIB with unscaled, raw bits & send to pvProcessBitmapAlpha
        aSource(5) = pvCreateDibDC(hDIB, aSource(1), aSource(2), 32)
        If hDIB = 0 Then GoTo exitRoutine
        pvConvertFormat gHandle, PixelFormat32bppRGB, aSource(5), -aSource(3)
        GdipDisposeImage gHandle: gHandle = 0: Erase aBits()
        Set oPic = pvHandleToPicture(hDIB, vbPicTypeBitmap)
        Set oPic = pvProcessBitmapAlpha(oPic, Width, Height, FillColor, mgmtStyle, alphaUsage, hAttrs, True)
        hDIB = 0: GoTo exitRoutine
    End If
    
    ' wanting icon/cursor handle, render the image
    
    If ICI(3) = 0 Then ' need a new mask
        ' create the mask vs cheating with an all-opaque mask, unpremultiply
        ReDim aData(0 To 5) ' get bits pointer
        GetGDIObject hDIB, 24, aData(0)
        mScanWidth = ((Width + &HF&) And Not &HF&) \ &H8& ' mask scanwdith
        ReDim aMask(0 To mScanWidth * Height - 1)
        ' overlay array on DIB bits
        pvSetOverlay uSafeArray(), VarPtrArray(aBits), aData(5), Width * Height
        Erase aData()
        For y = 0 To Height - 1
            ' setup variables for each scanline
            ' CreateBitmap has bits in top-down order, iterate mask bottom-up
            mPtr = (Height - y - 1) * mScanWidth
            x = y * Width: mskShift = &H80: mValue = &HFF ' default no bits opaque
            For x = x To x + Width - 1
                If (aBits(x) And &HFF000000) = &HFF000000 Then
                    mValue = mValue Xor mskShift   ' set mask bit to 0
                End If
                If mskShift = 1 Then
                    aMask(mPtr) = mValue: mValue = &HFF
                    mskShift = &H80: mPtr = mPtr + 1
                Else
                    mskShift = mskShift \ 2
                End If
            Next
            If mskShift <> &H80 Then aMask(mPtr) = mValue
        Next ' done with the mask, release overlay & create mask bitmap
        pvSetOverlay uSafeArray(), VarPtrArray(aBits), 0, 0
        ICI(3) = CreateBitmap(Width, Height, 1, 1, aMask(0))
        Erase aMask()
        If ICI(3) = 0 Then GoTo exitRoutine
    End If
    ' create icon/cursor and scale/mirror cursor hotspot if needed
    ICI(4) = hDIB: hDIB = 0
    If ICI(0) <> 0 Then                 ' cursor
        xRatio = Width / aSource(1)     ' x scale ratio
        If (ICI(0) And 2) = 2 Then      ' mirrored x-axis
            ICI(1) = (aSource(1) - ICI(1)) * xRatio - 1
            ICI(0) = ICI(0) Xor 2
        Else
            ICI(1) = ICI(1) * xRatio    ' not mirrored
        End If
        xRatio = Height / aSource(2)    ' y scale ratio
        If ICI(0) = 4 Then              ' mirrored y-axis
            ICI(2) = (aSource(2) - ICI(2)) * xRatio - 1
            ICI(0) = 0
        Else
            ICI(2) = ICI(2) * xRatio    ' not mirrored
        End If
    End If
    hIcon = CreateIconIndirect(ICI(0))
    If tPic.Type = vbPicTypeIcon Then
        Set oPic = pvHandleToPicture(hIcon, vbPicTypeIcon)
    End If
    
exitRoutine:    ' clean up
    If ICI(3) <> 0 Then DeleteObject ICI(3)
    If ICI(4) <> 0 Then DeleteObject ICI(4)
    If hDIB <> 0 Then DeleteObject hDIB
    If gHandle <> 0 Then GdipDisposeImage gHandle
    If oPic Is Nothing Then Set oPic = tPic
    Set pvProcessHICON = oPic
    
End Function

Private Function pvSetHGraphics(hDC As Long, hObject As Long, _
                                FillColor As Long, hAttrs As Long, _
                                FillOnly As Boolean, _
                                Optional modeGDI As Boolean = True) As Long

    ' helper function for setting up GDI+ hGraphics object
    ' if function returns zero, hGraphics failed creation else hGraphics returned
    
    ' modeGDI = True
    '   hDC must be passed, hObject (DIB) is optional
    '   if hObject passed and function succeeds then the hDib is selected into
    '       passed hDC. Before exiting calling routine, unselect:
    '       hDib = SelectObject(hDC, hDib)
    
    ' modeGDI = False
    '   hObject must be passed as GDI+ handle

    Dim tDC As Long, tObj As Long, hGraphics As Long
    Const HighQualityBiCubic As Long = 7
    Const PixelOffsetModeHalf = 4&
    
    If modeGDI = True Then
        If hObject <> 0 Then tObj = SelectObject(hDC, hObject)
        If GdipCreateFromHDC(hDC, hGraphics) <> 0 Then
            SelectObject hDC, tObj
            Exit Function
        End If
    ElseIf GdipGetImageGraphicsContext(hObject, hGraphics) <> 0 Then
        Exit Function
    End If
    If FillColor = -1 Then
        GdipGraphicsClear hGraphics, 0
    Else
        GdipGraphicsClear hGraphics, &HFF000000 Or (FillColor And &HFF) * &H10000 _
                    Or (FillColor And &HFF0000) \ &H10000 Or (FillColor And &HFF00&)
    End If
    If FillOnly = False Then
        If hAttrs = 0 Then
            If GdipCreateImageAttributes(hAttrs) = 0 Then
                GdipSetImageAttributesWrapMode hAttrs, WrapModeTileFlipXY, 0, 0
            End If
        End If
        GdipSetInterpolationMode hGraphics, HighQualityBiCubic
        GdipSetPixelOffsetMode hGraphics, PixelOffsetModeHalf
    End If
    If modeGDI = True And tObj <> 0 Then hObject = tObj
    pvSetHGraphics = hGraphics

End Function

Private Function pvDIBFromGDIplus(gHandle As Long, Width As Long, Height As Long, _
                    FillColor As Long, Optional pBits As Long = 0, _
                    Optional ByVal hAttrs As Long = 0, _
                    Optional sourceDimensions As Variant, _
                    Optional gdipTypeData As Long = 0) As Long
                    
    ' converts GDI+ handle to DIB while resizing, mirroring, filling alpha
    ' or renders from GDI+ to GDI+ if gdipTypeData is passed
    '   in that case, destination GDI+ handle returned as pBits
    
    ' if mirroring, Width & Height are non-negative on return
                    
    Dim hDIB As Long, hDC As Long, hGraphics As Long, gAttrs As Long
    Dim x As Long, y As Long, fData() As Single, gDestination As Long
    Dim Cx As Long, Cy As Long, lDepth As Long, bOk As Boolean
    Dim hBrush As Long
    
    If FillColor <> -1 Then
        lDepth = 24
    ElseIf hAttrs <> 0 Then
        lDepth = 32
    End If
    gAttrs = hAttrs
    Cx = Abs(Width): Cy = Abs(Height)
    
    If gdipTypeData = 0 Then        ' render GDI+ >> GDI hDC
        If lDepth = 0 Then lDepth = 32
        pBits = pvCreateDibDC(hDIB, Cx, Cy, lDepth, hDC)
        If hDIB = 0 Then Exit Function
        hGraphics = pvSetHGraphics(hDC, hDIB, FillColor, hAttrs, False)
    Else                            ' render GDI+ >> GDI+
        If gdipTypeData = -PixelFormat32bppARGB Then
            gdipTypeData = PixelFormat32bppARGB
        ElseIf lDepth = 0 Then
            Select Case (gdipTypeData And &HFF00&) \ &H100
                Case Is > 32: gdipTypeData = PixelFormat32bppARGB
                Case Is < 24: gdipTypeData = PixelFormat24bppRGB
            End Select
        ElseIf lDepth = 24 Then
            gdipTypeData = PixelFormat24bppRGB
        ElseIf hAttrs <> 0 Then
            gdipTypeData = PixelFormat32bppARGB
        End If
        lDepth = (gdipTypeData And &HFF00&) \ &H100
        pBits = pvCreateDibDC(hDIB, Cx, Cy, lDepth)
        If hDIB = 0 Then Exit Function
        y = (((Cx * lDepth) + &H1F&) And Not &H1F&) \ &H8&
        x = ((pBits Xor &H80000000) + y * (Cy - 1)) Xor &H80000000
        GdipCreateBitmapFromScan0 Cx, Cy, -y, gdipTypeData, ByVal x, gDestination
        If gDestination = 0 Then GoTo exitRoutine
        hGraphics = pvSetHGraphics(0&, gDestination, FillColor, hAttrs, False, False)
    End If
    If hGraphics = 0 Then GoTo exitRoutine
    If IsMissing(sourceDimensions) Then
        ReDim fData(0 To 3)
        GdipGetImageBounds gHandle, fData(0), x
    Else
        fData() = sourceDimensions
    End If
    If Width < 0 Then x = Cx Else x = 0
    If Height < 0 Then y = Cy Else y = 0
    GdipDrawImageRectRectI hGraphics, gHandle, x, y, Width, Height, fData(0), fData(1), fData(2), fData(3), UnitPixels, hAttrs, 0, 0
    GdipDeleteGraphics hGraphics
    bOk = True
    
exitRoutine:
    If hAttrs <> 0 And hAttrs <> gAttrs Then GdipDisposeImageAttributes hAttrs
    If hDC <> 0 Then
        hDIB = SelectObject(hDC, hDIB)
        DeleteDC hDC
        If bOk = False Then DeleteObject hDIB: hDIB = 0: pBits = 0
    ElseIf bOk = True Then
        pBits = gDestination
    Else
        If gDestination <> 0 Then GdipDisposeImage gDestination
        DeleteObject hDIB: hDIB = 0: pBits = 0
    End If
    If Width < 0 Then Width = Cx
    If Height < 0 Then Height = Cy
    pvDIBFromGDIplus = hDIB
    
End Function

Private Function pvDIBFromStdPicture(tPic As StdPicture, Width As Long, Height As Long, _
                    FillColor As Long, Optional pBits As Long) As Long
                    
    ' converts VB stdPicture GIF/Metafile to DIB
    ' if mirroring, Width & Height are non-negative on return
                    
    Dim hDIB As Long, hDC As Long, hGraphics As Long
    Dim x As Long, y As Long
    
    pBits = pvCreateDibDC(hDIB, Abs(Width), Abs(Height), 32, hDC)
    If hDIB = 0 Then Exit Function
    hGraphics = pvSetHGraphics(hDC, hDIB, FillColor, 0&, True)
    If hGraphics = 0 Then
        DeleteObject hDIB: hDIB = 0: pBits = 0
    Else
        GdipDeleteGraphics hGraphics
        If Width < 0 Then x = -Width
        If Height < 0 Then y = -Height
        With tPic
            .Render (hDC), (x), (y), (Width), (Height), 0, .Height, .Width, -.Height, ByVal 0&
        End With
        hDIB = SelectObject(hDC, hDIB)
    End If
    DeleteDC hDC
    If Width < 0 Then Width = -Width
    If Height < 0 Then Height = -Height
    pvDIBFromStdPicture = hDIB

End Function

Private Function pvCreateDibDC(hDIB As Long, ByVal Width As Long, ByVal Height As Long, _
                                ByVal Depth As Long, Optional hDC As Variant) As Long
    
    ' common routine for create 32bpp DIB & optional hDC
    '   DIB bits pointer returned by function, hDC returned if passed
    
    Dim aBMI(0 To 9) As Long, tDC As Long, dDC As Long
    
    hDIB = 0: tDC = GetDC(0)
    If tDC <> 0 Then
        If IsMissing(hDC) = False Then
            dDC = CreateCompatibleDC(tDC)
            If dDC = 0 Then GoTo exitRoutine
            hDC = dDC
        End If
    End If
    If Width <> 0 And Height <> 0 Then
        aBMI(0) = 40: aBMI(1) = Width: aBMI(2) = Height
        aBMI(3) = Depth * &H10000 Or 1
        hDIB = CreateDIBSection(tDC, aBMI(0), 0, pvCreateDibDC, 0, 0)
        If hDIB = 0 Then
            If hDC <> 0 Then DeleteDC hDC: hDC = 0
        End If
    End If
exitRoutine:
    ReleaseDC 0, tDC

End Function

Private Function pvStreamToPicture(oStream As stdole.IUnknown) As StdPicture
    ' create VB picture from stream
    Dim aGUID(0 To 3) As Long, oPic As IPicture
    
    aGUID(0) = &H7BF80980 ' GUID for IPicture {7BF80980-BF32-101A-8BBB-00AA00300CAB}
    aGUID(1) = &H101ABF32: aGUID(2) = &HAA00BB8B: aGUID(3) = &HAB0C3000
    ' create stdPicture
    OleLoadPicture ByVal ObjPtr(oStream), 0&, 0&, aGUID(0), oPic
    If oPic Is Nothing Then
        IStream_Reset oStream
    Else
        Set pvStreamToPicture = oPic
    End If

End Function

Private Function pvHandleToPicture(hHandle As Long, lType As PictureTypeConstants) As IPicture
    ' create VB picture from handle
    Dim aGUID(0 To 7) As Long
    
    aGUID(0) = 16: aGUID(1) = lType: aGUID(2) = hHandle: aGUID(3) = 0
    ' IPicture GUID {7BF80980-BF32-101A-8BBB-00AA00300CAB}
    aGUID(4) = &H7BF80980: aGUID(5) = &H101ABF32: aGUID(6) = &HAA00BB8B: aGUID(7) = &HAB0C3000
    ' create stdPicture
    OleCreatePictureIndirect aGUID(0), aGUID(4), True, pvHandleToPicture

End Function

Private Function pvGetFramePageCount(hHandle As Long, lType As ePicType, ByVal Index As Long) As Long
    ' return GIF/TIFF frame/page count and/or set current frame/page
    Dim aGUID(0 To 3) As Long, lCount As Long
    
    If lType = typeGIF Then
        ' GIFTimeDimension GUID {6AEDBD6D-3FB5-418A-83A6-7F45229DC872}
        aGUID(0) = &H6AEDBD6D: aGUID(1) = &H418A3FB5: aGUID(2) = &H457FA683: aGUID(3) = &H72C89D22
    Else
        ' TIFFPageDimension {7462DC86-6180-4C7E-8E3F-EE7333A7A483}
        aGUID(0) = &H7462DC86: aGUID(1) = &H4C7E6180: aGUID(2) = &H73EE3F8E: aGUID(3) = &H83A4A733
    End If
    GdipImageGetFrameCount hHandle, aGUID(0), lCount
    If lCount = 0 Then lCount = 1 Else lCount = lCount And &HFFF ' max supported
    If Index > 1 And lCount > 1 Then
        If Index > lCount Then Index = lCount
        GdipImageSelectActiveFrame hHandle, aGUID(0), Index - 1
    End If
    pvGetFramePageCount = lCount

End Function

Private Function pvStreamFromArray(aData() As Byte, _
                    Optional lOffset As Long = 0, Optional lSize As Long) As stdole.IUnknown
    ' creates an IStream compatible object from an array of bytes
    
    Dim hStream As Long, hLock As Long
   
    If lSize = 0 Then
        lOffset = LBound(aData)
        lSize = Abs(UBound(aData) - lOffset) + 1
    End If
    hStream = GlobalAlloc(&H2&, lSize)
    If hStream <> 0 Then
        hLock = GlobalLock(hStream)
        If hLock = 0 Then
            GlobalFree hStream
        Else
            CopyMemory ByVal hLock, aData(lOffset), lSize
            GlobalUnlock hStream
            CreateStreamOnHGlobal hStream, 1&, pvStreamFromArray
        End If
    End If
End Function

Private Sub pvSetOverlay(uSafeArray() As Long, arrayPtr As Long, pvData As Long, _
                            arraySize As Long, Optional elementSize As Long = 4)
    ' helper function to place a 1D zero-bound array on a memory address
    ' prevents having to do a ton of pointer math
    
    ' uSafeArray() = any Long array to be used as a SafeArray
    ' arrayPtr = VarPtr(array for overlaying)
    ' pvData = memory address to overlay
    ' arraySize = size in relation to element size; i.e. nr Longs, nr Bytes, etc
    ' elementSize = bytes per element (4=long, 2=integer, 1=byte)
    If pvData = 0 Then
        CopyMemory ByVal arrayPtr, 0&, 4
        Erase uSafeArray()
    Else
        ReDim uSafeArray(0 To 5)
        uSafeArray(0) = 1: uSafeArray(1) = elementSize
        uSafeArray(3) = pvData: uSafeArray(4) = arraySize
        CopyMemory ByVal arrayPtr, VarPtr(uSafeArray(0)), 4
    End If
    
End Sub

Private Function pvGuessImageType(Source() As Byte) As ePicType

' Purpose: Try to determine which image class the source data should be sent to for processing
'   Avoid brute force method of trying every format until one says it was processed
'   Note: the file extension (if loaded by file) is never assumed correct, not even tested
' Return: the image format type
    
' Image format magic numbers. Not all formats supported by this specific class
' (p) placeable WMF vs. (n) non-placeable WMF
' if byte column not populated, value can be anything or doesn't apply to magic number
'    Format  byte 1  byte 2  byte 3    byte 4  byte 9  byte 10 byte 11 byte 12     Magic.Nr.Bytes
'    --------------------------------------------------------------------------------------------------
'    ANI     82      73      70        70       65      67      79      78         12 (RIFF + 4bytes + ACON)
'    AVI     82      73      70        70       65      86      73      32         12 (RIFF + 4bytes + AVI )
'    BMP     66      77                                                            2 (BM)
'    CUR     0       0       2         0                                           none
'    EMF     0       1                                                             none
'    EMF+    70      77      69        32                                          4
'    GIF     71      73      70                                                    6 (GIF89a,GIF87a)
'    ICO     0       0       1         0                                           none
'    JPG     255     216                                                           2 (SOI tag)
'    MP3     73      68      51                                                    3 (ID3)
'    PCX     10      0,1     2,3,5                                                 none
'    PNG     137     80      78        71                                          8
'    PNM     80      1-7                                                           none
'    SWF     70      87      83                                                    3 (SWF)
'    TGA     0-255   0,1     0-3,9-11,32-33                                        none
'    TIFF    73,77   73,77                                                         4 (0x4949,0x4D4D)
'    WMA     48      38      117       0        166     217     0       170        16 (GUID)
'    WMF (n)
'    WMF (p) 215     205     198       154                                         4
'    *Binary 77      90                                                            2 (MZ)
' loading directly with GDI+ is the fallback procedure should source not be determined herein

    ' We target the 2nd byte as it is more unique than the 1st byte.
    ' Remaining bytes not always used for ID
    ' Any of the bytes can be examined and tested as needed

    Dim dWord As Long, byte1 As Byte
    On Error Resume Next
    byte1 = Source(LBound(Source))
    If Err Then Exit Function
    If Abs(UBound(Source) - LBound(Source)) < 32 Then Exit Function
    On Error GoTo 0
    
    CopyMemory dWord, Source(LBound(Source)), 4
    Select Case Source(LBound(Source) + 1) ' test 2nd byte
        ' looking in order of most popular file types
        Case 77     ' BMP,TIF,EMF+ testing
            Select Case byte1
            Case 66: pvGuessImageType = typeBitmap
            Case 77: pvGuessImageType = typeTIFF
            Case 70
                If dWord = &H20454D46 Then pvGuessImageType = typeMetafile
            End Select
        Case 216    ' JPG testing
            If byte1 = 255 Then pvGuessImageType = typeJPG
        Case 80     ' PNG testing
            If dWord = &H474E5089 Then pvGuessImageType = typePNG
        Case 73     ' GIF,TIFF,AVI,ANI testing
            Select Case byte1           '  AVI/ANI are RIFF formats
            Case 71: pvGuessImageType = typeGIF
            Case 73: pvGuessImageType = typeTIFF
            End Select
        Case 205    ' WMF (placeable) testing
            If dWord = &H9AC6CDD7 Then pvGuessImageType = typeMetafile
        Case 0      ' likely ICO/CUR,TGA,PCX
            If byte1 = 0 Then
                If (dWord And &HFF0000) = &H10000 Then
                    pvGuessImageType = typeIcon
                ElseIf (dWord And &HFF0000) = &H20000 Then
                    pvGuessImageType = typeCursor
                End If
            ElseIf byte1 = 1 Then
                pvGuessImageType = typeMetafile
            End If
        Case 1      ' EMF testing
            If byte1 = 0 Then pvGuessImageType = typeMetafile
    End Select
    
End Function

Private Function pvGetSysDPI(Optional ScreenDepth As Variant) As Long

    ' returns system DPI and optionally screen color depth
    Dim hDC As Long
    Const BITSPIXEL As Long = 12
    Const LOGPIXELSX As Long = 88
    hDC = GetDC(0)
    pvGetSysDPI = GetDeviceCaps(hDC, LOGPIXELSX)
    If IsMissing(ScreenDepth) = False Then ScreenDepth = GetDeviceCaps(hDC, BITSPIXEL)
    ReleaseDC 0, hDC
    
End Function

Private Function pvGetFileData(ByVal FileName As String, outArray() As Byte) As Boolean

    ' Function uses APIs to read files with unicode support

    Const GENERIC_READ As Long = &H80000000
    Const OPEN_EXISTING = &H3
    Const FILE_SHARE_READ = &H1
    Const FILE_ATTRIBUTE_ARCHIVE As Long = &H20
    Const FILE_ATTRIBUTE_HIDDEN As Long = &H2
    Const FILE_ATTRIBUTE_READONLY As Long = &H1
    Const FILE_ATTRIBUTE_SYSTEM As Long = &H4
    Const FILE_ATTRIBUTE_NORMAL As Long = &H80
    Const INVALID_HANDLE As Long = -1
    
    Dim Flags As Long, Access As Long, hHandle As Long
    Dim Disposition As Long, Share As Long
    
    Access = GENERIC_READ
    Share = FILE_SHARE_READ
    Disposition = OPEN_EXISTING
    Flags = FILE_ATTRIBUTE_ARCHIVE Or FILE_ATTRIBUTE_HIDDEN Or FILE_ATTRIBUTE_NORMAL _
            Or FILE_ATTRIBUTE_READONLY Or FILE_ATTRIBUTE_SYSTEM
    
    hHandle = CreateFileW(StrPtr(FileName), Access, Share, ByVal 0&, Disposition, Flags, 0&)
    If hHandle = INVALID_HANDLE Then hHandle = 0
    If hHandle <> 0 Then
        Flags = GetFileSize(hHandle, ByVal 0&)
        If Flags > 0 Then
            ReDim outArray(0 To Flags - 1)
            ReadFile hHandle, outArray(0), Flags, Access, ByVal 0&
            pvGetFileData = True
        End If
        CloseHandle hHandle
    End If
    
End Function

Private Function pvReverseLong(ByVal inLong As Long) As Long

    ' fast function to reverse a long value from big endian to little endian
    ' PNG files contain reversed longs, as do ID3 v3,4 tags
    pvReverseLong = _
      (((inLong And &HFF000000) \ &H1000000) And &HFF&) Or _
      ((inLong And &HFF0000) \ &H100&) Or _
      ((inLong And &HFF00&) * &H100&) Or _
      ((inLong And &H7F&) * &H1000000)
    If (inLong And &H80&) Then pvReverseLong = pvReverseLong Or &H80000000
    
End Function

Private Function pvIsPictureValid(thePicture As StdPicture) As Boolean
    ' simply test picture for Nothing and/or null handle
    If Not thePicture Is Nothing Then pvIsPictureValid = (thePicture.Handle <> 0)
End Function

Private Function pvValidateAlphaChannel(pBits As Long, Size As Long) As eAlphaType
       
    ' Method determines if the alhpa channel is used and how it is used
    ' Speedy? For complex/standard transparency, should exit routine very fast.
    '   For anything else, all pixels will be scanned unfortuntely.
    '   Just can't exit until format is known for sure.
    
    ' Routine pulled from my AlphaImage control & slightly tweaked
    ' 4 Alpha formats are tracked by this routine
    '   Alpha channel not used. If all alpha bytes are just 0 or just 255
    '       note: all zero alpha bytes are assumed to be no alpha channel in use; otherwise 100% transparent
    '   Simple Transparency: all alpha bytes are a mix of only 0 or 255
    '   Complex Transparency: any alpha byte has range 1 to 254 and R,G,B channel > alpha channel
    '   PreMultiplied RGB Transparency: all R,G,B channel values <= alpha channel
   
    Dim bAlpha As Long, lFormat As Long
    Dim lColor As Long, lPrevColor As Long, x As Long
    Dim uSafeArray() As Long, arrBits() As Long
    
    Const PARGB As Long = 16
    Const ARGB As Long = 8
    Const Simple255 As Long = 2
    Const Simple0 As Long = 1
    Const ZEROES As Long = &H100&

    On Error GoTo exitRoutine
    pvSetOverlay uSafeArray(), VarPtrArray(arrBits), pBits, Size
    ' ///// test the alpha channel. Loop aborts early if it can
    lPrevColor = arrBits(0) Xor 1&
    For x = 0& To UBound(arrBits)
        lColor = arrBits(x)                 ' get 32bit color
        If lColor <> lPrevColor Then        ' and extact the alpha byte
            If lColor = 0& Then
                lFormat = lFormat Or ZEROES ' entire value is zero
                ' all zeroes indicates 100% transparent or 100% black image
                ' mix of zero & non-zero alpha values indicates transparency
            Else
                bAlpha = (lColor And &HFF000000) \ &H1000000 And &HFF
                If bAlpha = 0& Then ' RGB value is non-zero
                    If (lFormat And &HFF) > Simple255 Then
                        ' at least one other alpha value was > 0 and < 255
                        ' since this alpha is zero & RGB non-zero. Done
                        lFormat = ARGB: Exit For
                    End If
                    lFormat = lFormat Or Simple0 ' keep going, maybe all alphas are zero
                ElseIf bAlpha = 255& Then
                    If (lFormat And Simple0) = Simple0 Then
                        ' already seen alpha zero & non-zero RGB. Here we have 255 alpha. Done:
                        lFormat = ARGB: Exit For
                    End If
                    lFormat = lFormat Or Simple255
                   
                ' else if any RGB values > alpha then not-premultiplied
                ElseIf bAlpha < (lColor And &HFF&) Then
                    lFormat = ARGB: Exit For ' definitly ARGB
                ElseIf bAlpha < (lColor And &HFF00&) \ &H100& Then
                    lFormat = ARGB: Exit For ' definitly ARGB
                ElseIf bAlpha < (lColor And &HFF0000) \ &H10000 Then
                    lFormat = ARGB: Exit For ' definitly ARGB
                Else
                    lFormat = lFormat Or PARGB ' likely pARGB, but not sure yet
                End If
            End If
            lPrevColor = lColor
        End If
    Next
    pvSetOverlay uSafeArray(), VarPtrArray(arrBits), 0, 0
    
    If (lFormat And &HFF) = ARGB Then     ' alpha, not premultiplied
        pvValidateAlphaChannel = PixelFormat32bppARGB
    ElseIf lFormat = (ZEROES Or Simple255) Then ' simple transparency
        ' this is both ARGB & pARGB - set to custom pixel format
        ' all transparent pixels are zero alpha + zero RGB
        pvValidateAlphaChannel = PixelFormat32bppUARGB
    ElseIf (lFormat And PARGB) = PARGB Then     ' premultiplied RGB
        pvValidateAlphaChannel = PixelFormat32bppPARGB
    End If                                  ' else assume alpha channel not used
   
exitRoutine:
End Function

Private Function pvCreateToken(hToken As Long, ByVal asNeeded As Boolean) As Boolean
    ' simply creates a GDI+ token, as needed
    hToken = 0
    If asNeeded = True And Not m_Canary Is Nothing And m_GdipToken <> 0 Then
        pvCreateToken = True
    Else
        Dim aData(0 To 3) As Long, lToken As Long
        aData(0) = 1: GdiplusStartup lToken, aData(0)
        If asNeeded = False Then hToken = lToken Else m_GdipToken = lToken
        pvCreateToken = (lToken <> 0)
    End If
End Function

Private Function pvIsManaged(tPic As IPicture, Optional hClient As Long, _
                   Optional gdipHandle As Long, Optional hThunk As Long, _
                   Optional hClients As Long) As Long
                   
    #If AllowThunkCapability = True Then ' else all code below not compiled
    
    ' simply returns whether or not passed picture is managed by class thunk
    ' if no addresses need to be returned, use pvGetCanary instead
    
    ' hClient: start in hClients where this picture's record starts
    ' hClients: start of 4 entry record for passed picture
    ' hThunk: the thunk
    ' gdipHandle: GDI+ handle used for rendering
    ' Return value: data array address

    If pvIsPictureValid(tPic) = False Then Exit Function
    
    Dim lPtr As Long, uSafeArray() As Long
    Dim aData() As Long, n As Long
    Dim oCanary As stdole.IUnknown
    
    Set oCanary = pvGetCanary(tPic) ' managed pictures will return its canary
    If oCanary Is Nothing Then Exit Function
    
    CopyMemory hClients, ByVal ((ObjPtr(oCanary) Xor &H80000000) + Offset_ClientList) Xor &H80000000, 4
    ' overlay array onto thunk's client array
    CopyMemory n, ByVal hClients, 4
    pvSetOverlay uSafeArray(), VarPtrArray(aData), hClients, n
    
    ' see pvHookPicture for details regarding the thunk client record structure
    lPtr = ObjPtr(tPic)                     ' search criteria
    For n = 3 To uSafeArray(4) - 1 Step 4   ' 1st client starts @ offset 12 bytes
        If aData(n) = lPtr Then
            hClient = VarPtr(aData(n))
            gdipHandle = aData(n + 3)   ' indicates if 32bppPARGB
            pvIsManaged = aData(n + 2)
            Exit For
        End If
    Next
    hThunk = aData(2)
    pvSetOverlay uSafeArray(), VarPtrArray(aData), 0, 0
    Set oCanary = Nothing
    
    #End If
    
End Function

Private Function pvCanaryExists() As Boolean

    #If AllowThunkCapability = True Then ' else all code below not compiled

    ' asks caller for canary if one not attached to the class
    If m_Canary Is Nothing Then
        RaiseEvent Attach
        pvCanaryExists = Not (m_Canary Is Nothing)
    Else
        pvCanaryExists = True
    End If
    
    #End If
    
End Function

Private Sub pvHookPicture(tPic As StdPicture, gHandle As Long, _
                        Optional picType As ePicType, _
                        Optional CustProp As Long = 0, Optional CustAttrs As Long = 0, _
                        Optional CustMisc As Long = 0)

    ' performs hooking of the passed stdPicture
    
    #If AllowThunkCapability = True Then    ' else all code below not compiled
    
    If gHandle = 0 Then Exit Sub
    If m_Canary Is Nothing Then GoTo exitRoutine

    ' The thunk's client array is formatted as 3 item header + 4 item records:
    '   entry #0    header 1: size of client array in Longs
    '   entry #1    header 2: client count (nr StdPictures hooked)
    '   entry #2    header 3: thunk address (used only by this class)
    '   entry #3    client 1: Original IPicture ObjPtr
    '   entry #4    client 1: Original StdPicture ObjPtr
    '   entry #5   *client 1: Data array address (see next section)
    '   entry #6    client 1: GDI+ image handle
    '   entry #7    starts client 2 info
    
    ' *Each client has own Data array having 30 entries (25 are VTable entries)
    '  -- entries 2,3,4 are the 3 passed Cust[xxx] parameters
    '   entry #0    original VTable address for IPicture, for unhooking
    '   entry #1    original VTable address for StdPicture, for unhooking
    '   entry #2    Flags, currently used only for GIF with frame-navigation enabled
    '       0x80000000 - navigation enabled
    '       0x30000000 - all formats: 2=bicubic scaling,1=bilinear,0=nearest-neighbor
    '       0x0FFF0FFF - frame count (hiword), frame index (loword)
    '       0x00007000 - bit 1=render resized,2=mirrored-x,4=mirrored-y
    '       0x40008000 - reserved
    '   entry #3    User-supplied GDI+ image attributes
    '   entry #4    reserved
    '   entry #5-11  StdPicture's 7 VTable entries with IUnk::Release redirected
    '   entry #12-29 IPicture's 18 VTable entries with IUnk:QI, IUnk:Release,
    '                              IPic:Render, IPic:GetAttrs & IPic:hPal redirected

    Dim oPic As IPicture, hThunk As Long, hNewAddr As Long
    Dim lSize As Long, n As Long, hData As Long
    Dim hClients As Long, objSlotAddr As Long
    Dim aData() As Long, uSafeArray() As Long ' FAUX SafeArray structure
    Const CLIENTDATA_SIZE As Long = 30
    Const IPIC_Render As Long = &H98        ' thunk offsets for hooks
    Const IPIC_Attrs As Long = &H90
    Const IPIC_QI As Long = &H80
    Const IPIC_Release As Long = &H88
    Const IPIC_hPalGet As Long = &HA0
    Const IPIC_hPalLet As Long = &HA4
    Const IPIC_RedirectOffset = 48&         ' redirection offset into client array
    Const IDISP_RedirectOffset = 20&        ' redirection offset into client array

    hData = CoTaskMemAlloc(CLIENTDATA_SIZE * 4)
    If hData = 0 Then GoTo exitRoutine
    
    objSlotAddr = ((ObjPtr(m_Canary) Xor &H80000000) + Offset_ClientList) Xor &H80000000
    ' overlay array onto thunk's client array
    CopyMemory hClients, ByVal objSlotAddr, 4
    CopyMemory lSize, ByVal hClients, 4
    pvSetOverlay uSafeArray(), VarPtrArray(aData), hClients, lSize
    hThunk = aData(2)
    ' see if an open slot exists, search for IPicture ObjPtr, not StdPicture ObjPrr
    If aData(1) < aData(0) \ 4 Then ' open slot exists, find it
        For n = 3 To uSafeArray(4) - 1 Step 4
            If aData(n) = 0 Then Exit For
        Next
    Else    ' need to resize the array
        n = uSafeArray(4): lSize = n * 4 + 64         ' resize with 4 extra records
        pvSetOverlay uSafeArray(), VarPtrArray(aData), 0, 0
        ReDim aData(0 To lSize \ 4 - 1)
        CopyMemory aData(0), ByVal hClients, lSize - 64
        aData(0) = aData(0) + 16             ' increment nr of array items
        hNewAddr = CoTaskMemAlloc(lSize)
        If hNewAddr = 0 Then GoTo exitRoutine ' failed to resize array
        CopyMemory ByVal hNewAddr, aData(0), lSize
        CopyMemory ByVal objSlotAddr, hNewAddr, 4
        CoTaskMemFree hClients
        lSize = aData(0): Erase aData()
        pvSetOverlay uSafeArray(), VarPtrArray(aData), hNewAddr, lSize
    End If
    Set oPic = tPic
    aData(1) = aData(1) + 1             ' increment client count
    aData(n) = ObjPtr(oPic)             ' cache ObjPtrs
    aData(n + 1) = ObjPtr(tPic)
    aData(n + 2) = hData                ' cache client's data array address
    aData(n + 3) = gHandle              ' cache GDI+ handle, if any
    pvSetOverlay uSafeArray(), VarPtrArray(aData), 0, 0
    
    ' update the client's data array now
    pvSetOverlay uSafeArray(), VarPtrArray(aData), hData, CLIENTDATA_SIZE
    CopyMemory aData(0), ByVal ObjPtr(oPic), 4   ' original VTable address
    CopyMemory aData(1), ByVal ObjPtr(tPic), 4   ' original VTable address
    aData(2) = CustProp Or sqBiCubic * &H10000000 ' Flags + default BiCubic scaling
    aData(3) = CustAttrs                ' user-supplied ImageAttributes
    aData(4) = CustMisc                 ' reserved/miscellaneous as of this version
    CopyMemory aData(5), ByVal aData(1), 28   ' 7 table entries
    CopyMemory aData(12), ByVal aData(0), 72  ' 18 table entries
    ' of the 25 copied VTable entries, change these five to reroute to the thunk
    aData(7) = ((hThunk Xor &H80000000) + IPIC_Release) Xor &H80000000  ' tPic:Release
    aData(12) = ((hThunk Xor &H80000000) + IPIC_QI) Xor &H80000000      ' oPic:QueryInterface
    aData(14) = ((hThunk Xor &H80000000) + IPIC_Release) Xor &H80000000 ' oPic:Release
    aData(28) = ((hThunk Xor &H80000000) + IPIC_Attrs) Xor &H80000000   ' oPic:GetAttrs
    aData(20) = ((hThunk Xor &H80000000) + IPIC_Render) Xor &H80000000  ' oPic:Render
    If picType = typeGIF Then       ' reroute hPal only for GIFs
        aData(16) = ((hThunk Xor &H80000000) + IPIC_hPalGet) Xor &H80000000  ' oPic:hPalGet
        aData(21) = ((hThunk Xor &H80000000) + IPIC_hPalLet) Xor &H80000000  ' oPic:hPalLet
    End If
    pvSetOverlay uSafeArray(), VarPtrArray(aData), 0, 0
    
    ' VTable redirection for tPic starts at client data array's 16th byte offset
    CopyMemory ByVal ObjPtr(tPic), ((hData Xor &H80000000) + IDISP_RedirectOffset) Xor &H80000000, 4
    ' VTable redirection for oPic starts at client data array's 44th byte offset
    CopyMemory ByVal ObjPtr(oPic), ((hData Xor &H80000000) + IPIC_RedirectOffset) Xor &H80000000, 4
    Set oPic = Nothing
    Exit Sub


exitRoutine:    ' clean up
    If hData <> 0 Then CoTaskMemFree hData
    #End If
    
    If gHandle <> 0 Then GdipDisposeImage gHandle
End Sub

#If AllowThunkCapability = True Then
'////// 100% thunk-only related private methods

Private Function pvGetCanary(theObject As stdole.IUnknown) As stdole.IUnknown

    ' returns the canary object related to the passed object
    ' the passed object can be a stdPicture
    
    Dim aGUID(0 To 3) As Long, pParams(0 To 1) As Long
    Dim vParams(0 To 2) As Variant, pParamType(0 To 1) As Integer
    Dim oCanary As stdole.IUnknown

    ' custom GUID {5a752873-1908-42b0-1962-d1f14a3efa3a}
    aGUID(0) = &H5A752873: aGUID(1) = &H42B01908
    aGUID(2) = &HF1D16219: aGUID(3) = &H3AFA3E4A
    ' set up call for QueryInterface
    vParams(0) = VarPtr(aGUID(0)): vParams(1) = VarPtr(oCanary)
    pParams(0) = VarPtr(vParams(0)): pParamType(0) = vbLong
    pParams(1) = VarPtr(vParams(1)): pParamType(1) = vbLong
    DispCallFunc ObjPtr(theObject), 0, 4, vbLong, 2, VarPtr(pParamType(0)), VarPtr(pParams(0)), vParams(2)
    Set pvGetCanary = oCanary: Set oCanary = Nothing

End Function

Private Function pvScaleAnimatedGIF(thePicture As StdPicture, Width As Long, Height As Long, _
                                FillColor As Long, mgmtStyle As ManagedStyle, _
                                alphaUsage As AlphaUsageEnum, hAttrs As Long, forceNewBmp As Boolean) As StdPicture

    ' Called by Me.CopyPicture -- pre-validated as managed animated GIF

    ' Animated GIF (navigation enabled) is not truly scaled. Rather, it is
    ' rendered scaled from its original dimensions when it was first loaded.
    
    ' The thunk knows how to handle scaling/mirroring from a property stored
    ' with the picture's client data entries. When we scale/mirror the animated
    ' GIF, we update that property and create a DIB to be used as the
    ' placeholder for the picture, at scaled size. If the placeholder changes
    ' size, the existing placeholder is unhooked & new one is hooked in its place
    
    ' see pvHookPicture for details of the client data array
    
    Dim gHandle As Long, hClient As Long, hClients As Long
    Dim hAddr As Long, hThunk As Long, lValue As Long
    Dim hDIB As Long, gHookHandle As Long
    Dim lFlags As Long, fRect(0 To 3) As Single
    Dim aSource(0 To 5) As Long, oPic As StdPicture
    Dim oCanary As stdole.IUnknown
    Const Offset_GdipHandle As Long = 12
    
    hAddr = pvIsManaged(thePicture, hClient, gHandle, hThunk, hClients)
    GdipGetImageBounds gHandle, fRect(0), lValue
    
    ' handle management type changes, filling & attrs
    If forceNewBmp = False Then
        If FillColor <> -1 Or hAttrs <> 0 Then
            forceNewBmp = True
        ElseIf mgmtStyle <> msAutoSelect Or alphaUsage <> acApplyAlphaChannel Then
            forceNewBmp = True
        End If
    End If
    If forceNewBmp = True Then
        Set pvScaleAnimatedGIF = pvProcessBitmapAlpha(thePicture, Width, Height, FillColor, mgmtStyle, alphaUsage, hAttrs, False)
        GoTo exitRoutine
    End If
    
    GetGDIObject thePicture.Handle, 24, aSource(0)
    If Width = 0 Then               ' mirroring flags
        Width = aSource(1)          ' 1 = scaling
    ElseIf Width < 0 Then           ' 2 = mirror x axis
        lFlags = lFlags Or &H2000   ' 4 = mirror y axis
        Width = -Width              ' 6 = mirror mask
    End If
    If Height = 0 Then
        Height = aSource(2)
    ElseIf Height < 0 Then
        lFlags = lFlags Or &H4000
        Height = -Height
    End If
    
    ' get the GIF props and adjust based on passed params
    CopyMemory aSource(3), ByVal ((hAddr Xor &H80000000) + Offset_ClientPropA) Xor &H80000000, 12
    aSource(3) = ((aSource(3) And &HFFFFEFFF) Xor (lFlags And &H6000))
    If Abs(Width) <> fRect(2) Or Abs(Height) <> fRect(3) Then
        aSource(3) = aSource(3) Or &H1000   ' placeholder size different than GIF
    End If
    
    ' is the placeholder size changing?
    If Abs(Width) <> aSource(1) Or Abs(Height) <> aSource(2) Then
        If Not m_Canary Is pvGetCanary(thePicture) Then
            ' when rehooking, use the same canary
            Set oCanary = m_Canary
            Set m_Canary = pvGetCanary(thePicture)
        End If
        ' need to resize the placeholder VB picture
        hDIB = pvDIBFromGDIplus(gHandle, Abs(Width), Abs(Height), vbWhite, , , fRect())
        If hDIB = 0 Then GoTo exitRoutine
        Set oPic = pvHandleToPicture(hDIB, vbPicTypeBitmap)
        pvHookPicture oPic, gHandle, typeGIF, aSource(3), aSource(4), aSource(5)
        If Not oCanary Is Nothing Then
            Set m_Canary = oCanary
            Set oCanary = Nothing
        End If
        
        ' prepare & unhook passed picture; ensure attrs & gHandle not released
        Erase aSource()
        CopyMemory ByVal ((hAddr Xor &H80000000) + Offset_ClientPropA) Xor &H80000000, aSource(0), 12
        CopyMemory ByVal ((hClient Xor &H80000000) + Offset_GdipHandle) Xor &H80000000, 0&, 4
        CallThunkProc ((hThunk Xor &H80000000) + Offset_Unhook) Xor &H80000000, hThunk, hClients, hClient, 0
        
    Else ' update flags only
        CopyMemory ByVal ((hAddr Xor &H80000000) + Offset_ClientPropA) Xor &H80000000, aSource(3), 4
    End If
    
exitRoutine:
    If oPic Is Nothing Then Set oPic = thePicture
    Set pvScaleAnimatedGIF = oPic
    
End Function

Private Function pvCreateThunk(hCanary As Long, pUnkQI As Long, _
                        pUnkAddRef As Long, pUnkRelease As Long) As Long

    ' creates just one thunk used by the entire project
    ' all hooked StdPictures will be processed by this thunk
    
    ' The thunk itself is fairly small. It hooks up to 7 methods and has
    ' one custom method to be called only from this class.
    
    ' That custom method enables class UnmanagePicture to be called.
    ' The hooked picture methods are:
    '   Render: uses GDI+ to draw/scale image as needed, bypasses COM method
    '   Attributes Get: Ensures PICTURE_TRANSPARENT flag returned to VB
    '   IUnknown:Release: looks for the reference count reaching zero
    '       this is hooked both on the stdPicture & its IPicture interfaces
    '   IUnknown:QueryInterface: returns a managed picture's hook object
    '   hPal Get/Let: repurposed to optionally enable GIF frame navigation
    '       those last 2 methods only hooked for navigation-enabled GIFs
    
    ' notes to self: Any thunk modifications may result in needing to
    '   update offsets in pvHookPicture & class-level constants
    ' ensure #ThunkReleaseTesting is False before compiling into real application
    ' or before posting update to vbForums. When set to True, a MsgBox will be
    ' displayed indicating how many hooked pictures remain at the time the
    ' thunk is about to be released. This is for debugging only to ensure the
    ' count is correct (if applicable) and to ensure the thunk is self-releasing.
    ' When set to True, extra 160-ish bytes of code added to the thunk and a MsgBox
    ' will always be displayed -- i.e., testing purposes only!
    
    ' FYI for those that aren't up to speed with compiler directives: # usage
    '   When #IF is false, code enclosed in #IF statement isn't compiled into
    '   project at all. It is as if the source code doesn't even exist.
    
    Dim hToken As Long, uSafeArray() As Long
    Dim hMod As Long, hThunk As Long
    Dim z_Sc() As Long, aData() As Byte
    Const CB_PAGE_RWX    As Long = &H40     ' Allocate executable memory
    Const CB_MEM_COMMIT  As Long = &H1000   ' Commit allocated memory
    Const CanaryQI As Long = &HA8           ' 3 IUnknown offsets in thunk
    Const CanaryAddRef As Long = &HB8
    Const CanaryRelease As Long = &HB0
#If ThunkReleaseTesting = False Then        ' thunk size in Longs
    Const THUNK_SIZE As Long = 364          ' without debug code
#Else
    Const THUNK_SIZE As Long = 402          ' with debug code
#End If
    
    ' create a token for the thunk
    If pvCreateToken(hToken, False) = False Then Exit Function
    ' VirtualAlloc zeroizes allocation data
    hThunk = VirtualAlloc(0&, THUNK_SIZE * 4&, CB_MEM_COMMIT, CB_PAGE_RWX)
    If hThunk = 0 Then
        GdiplusShutdown hToken
        Exit Function
    End If
    
    ' overlay array on to allocation & create thunk
    pvSetOverlay uSafeArray(), VarPtrArray(z_Sc), hThunk, THUNK_SIZE
    z_Sc(0) = &H1000BB: z_Sc(1) = &HE800&: z_Sc(2) = &H815B0000: z_Sc(3) = &HAEB&: z_Sc(4) = &HF689C300
    z_Sc(30) = &H10FE9: z_Sc(31) = &H768D00: z_Sc(32) = &H33FE9: z_Sc(33) = &H768D00: z_Sc(34) = &H37FE9: z_Sc(35) = &H768D00: z_Sc(36) = &H303E9: z_Sc(37) = &H768D00: z_Sc(38) = &H107E9: z_Sc(39) = &H768D00: z_Sc(40) = &HF6895EEB: z_Sc(41) = &HF6897AEB: z_Sc(42) = &H43BE9: z_Sc(43) = &H768D00: z_Sc(44) = &H477E9: z_Sc(45) = &H768D00: z_Sc(46) = &H424548B: z_Sc(47) = &H4004428B: z_Sc(48) = &HC2044289: z_Sc(49) = &HF6890004: z_Sc(50) = &HE8E58960: z_Sc(51) = &HFFFFFF30: z_Sc(52) = &HFF2C75FF: z_Sc(53) = &H93FF2475: z_Sc(54) = &H28&: z_Sc(55) = &H8920458B: z_Sc(56) = &H5D892445: z_Sc(57) = &H2C45C728: z_Sc(59) = &H3045C7
    z_Sc(60) = &H8B000080: z_Sc(61) = &H6083&: z_Sc(62) = &H1C458900: z_Sc(63) = &HE0FF5961: z_Sc(64) = &H8960C031: z_Sc(65) = &HFEF6E8E5: z_Sc(66) = &H89E8FFFF: z_Sc(67) = &H8B000003: z_Sc(68) = &HFF250847: z_Sc(69) = &H8B0FFF0F: z_Sc(70) = &H2892855: z_Sc(71) = &H8C261: z_Sc(72) = &H8960C031: z_Sc(73) = &HFED6E8E5: z_Sc(74) = &H69E8FFFF: z_Sc(75) = &H8B000003: z_Sc(76) = &HC889084F: z_Sc(77) = &HE181&: z_Sc(78) = &H4CE38000: z_Sc(79) = &HFFF2550: z_Sc(80) = &HE3590000: z_Sc(81) = &HE9C15143: z_Sc(82) = &HFFE18110: z_Sc(83) = &H8B00000F: z_Sc(84) = &HE2812855: z_Sc(85) = &HFFFF&: z_Sc(86) = &H7C01FA83: z_Sc(87) = &H7ECA390B: z_Sc(88) = &H1BA09: z_Sc(89) = &H2EB0000
    z_Sc(90) = &H2558CA89: z_Sc(91) = &HFFFFF000: z_Sc(92) = &H4789D009: z_Sc(93) = &H89524A08: z_Sc(94) = &H6405D8: z_Sc(95) = &HFF500000: z_Sc(96) = &H93FF0C76: z_Sc(97) = &H50&: z_Sc(98) = &H8C261: z_Sc(99) = &H8BE58960: z_Sc(100) = &H7D8B245D: z_Sc(101) = &H2C758B28: z_Sc(102) = &H2A3E8: z_Sc(103) = &HD0896100: z_Sc(104) = &H900010C2: z_Sc(105) = &H8960C031: z_Sc(106) = &HFE52E8E5: z_Sc(107) = &H4539FFFF: z_Sc(108) = &H393A7444: z_Sc(109) = &H35744845: z_Sc(110) = &H74344539: z_Sc(111) = &H38453930: z_Sc(112) = &HEC832B74: z_Sc(113) = &H2CEE808: z_Sc(114) = &H478B0000: z_Sc(115) = &H25E85008: z_Sc(116) = &H59000000: z_Sc(117) = &HE181&: z_Sc(118) = &H5E38000: z_Sc(119) = &HCBE8&
    z_Sc(120) = &H5AE800: z_Sc(121) = &HC4830000: z_Sc(122) = &H2CC26108: z_Sc(123) = &H1C45C700: z_Sc(124) = &H80070057: z_Sc(125) = &HF689F3EB: z_Sc(126) = &HE844458B: z_Sc(127) = &H184&: z_Sc(128) = &H8B444589: z_Sc(129) = &HC0853C45: z_Sc(130) = &H75E80A74: z_Sc(131) = &H89000001: z_Sc(132) = &HC0313C45: z_Sc(133) = &HE848452B: z_Sc(134) = &H168&: z_Sc(135) = &H8D484589: z_Sc(136) = &HFF50FC45: z_Sc(137) = &H178B2475: z_Sc(138) = &H8B1C52FF: z_Sc(139) = &H452BFC45: z_Sc(140) = &H74C08540: z_Sc(141) = &H14AE805: z_Sc(142) = &H45890000: z_Sc(143) = &HF689C340: z_Sc(144) = &H50F8458D: z_Sc(145) = &HFF2875FF: z_Sc(146) = &H3C93&: z_Sc(147) = &H75C08500: z_Sc(148) = &H8478B59: z_Sc(149) = &H25&
    z_Sc(150) = &H1CE8C130: z_Sc(151) = &H5005C083: z_Sc(152) = &HFFF875FF: z_Sc(153) = &H3493&: z_Sc(154) = &HC478B00: z_Sc(155) = &H675C085: z_Sc(156) = &H20838B: z_Sc(157) = &H6A0000: z_Sc(158) = &H6A50006A: z_Sc(159) = &H4875FF02: z_Sc(160) = &HFF4475FF: z_Sc(161) = &H75FF4075: z_Sc(162) = &H3875FF3C: z_Sc(163) = &HFF3475FF: z_Sc(164) = &H75FF3075: z_Sc(165) = &HC76FF2C: z_Sc(166) = &HFFF875FF: z_Sc(167) = &H4493&: z_Sc(168) = &HF875FF00: z_Sc(169) = &H4093FF: z_Sc(170) = &H90C30000: z_Sc(171) = &H8108578B: z_Sc(172) = &H7000E2: z_Sc(173) = &HFD28500: z_Sc(174) = &HC684&: z_Sc(175) = &HE2815200: z_Sc(176) = &H1000&: z_Sc(177) = &H7A74D285: z_Sc(178) = &H50F8458D: z_Sc(179) = &HFF0C76FF
    z_Sc(180) = &H5893&: z_Sc(181) = &HFC458B00: z_Sc(182) = &HA7E8&: z_Sc(183) = &HF8558B00: z_Sc(184) = &H4D8B5250: z_Sc(185) = &H500CE340: z_Sc(186) = &H93FF5152: z_Sc(187) = &H5C&: z_Sc(188) = &HFF404589: z_Sc(189) = &H93FF4875: z_Sc(190) = &H5C&: z_Sc(191) = &H8D484589: z_Sc(192) = &HFF50F845: z_Sc(193) = &H93FF0C76: z_Sc(194) = &H54&: z_Sc(195) = &H50FC458D: z_Sc(196) = &H8B2475FF: z_Sc(197) = &H1852FF17: z_Sc(198) = &HE8FC458B: z_Sc(199) = &H64&: z_Sc(200) = &H50F8558B: z_Sc(201) = &H3C4D8B52: z_Sc(202) = &H52500CE3: z_Sc(203) = &H5C93FF51: z_Sc(204) = &H89000000: z_Sc(205) = &H75FF3C45: z_Sc(206) = &H5C93FF44: z_Sc(207) = &H89000000: z_Sc(208) = &H81594445: z_Sc(209) = &H6000E1
    z_Sc(210) = &H5138E300: z_Sc(211) = &H2000E181: z_Sc(212) = &H13E30000: z_Sc(213) = &H8B34458B: z_Sc(214) = &HC1012C4D: z_Sc(215) = &H312C4D89: z_Sc(216) = &H344D2BC9: z_Sc(217) = &H59344D89: z_Sc(218) = &H4000E181: z_Sc(219) = &H13E30000: z_Sc(220) = &H8B38458B: z_Sc(221) = &HC101304D: z_Sc(222) = &H31304D89: z_Sc(223) = &H384D2BC9: z_Sc(224) = &HC3384D89: z_Sc(225) = &H9EC68: z_Sc(226) = &H1CB3FF00: z_Sc(227) = &H50000000: z_Sc(228) = &H5C93FF: z_Sc(229) = &H90C30000: z_Sc(230) = &HE8E58960: z_Sc(231) = &HFFFFFC60: z_Sc(232) = &HF3E8&: z_Sc(233) = &HFF078B00: z_Sc(234) = &H75FF2875: z_Sc(235) = &H4050FF24: z_Sc(236) = &H851C4589: z_Sc(237) = &H8B0675C0: z_Sc(238) = &H983284D: z_Sc(239) = &H8C26102
    z_Sc(240) = &H768D00: z_Sc(241) = &HE8E58960: z_Sc(242) = &HFFFFFC34: z_Sc(243) = &HC7E8&: z_Sc(244) = &HFF078B00: z_Sc(245) = &H75FF2C75: z_Sc(246) = &H2475FF28: z_Sc(247) = &H458910FF: z_Sc(248) = &H74C0851C: z_Sc(249) = &H284D8B23: z_Sc(250) = &HD3E8&: z_Sc(251) = &H74C98500: z_Sc(252) = &H2C558B17: z_Sc(253) = &H14838B: z_Sc(254) = &H2890000: z_Sc(255) = &H4204508B: z_Sc(256) = &H31045089: z_Sc(257) = &H1C4589C0: z_Sc(258) = &HCC261: z_Sc(259) = &HE8E58960: z_Sc(260) = &HFFFFFBEC: z_Sc(261) = &H7FE8&: z_Sc(262) = &H14458900: z_Sc(263) = &H478B05E3: z_Sc(264) = &H8B02EB04: z_Sc(265) = &H2475FF07: z_Sc(266) = &H890850FF: z_Sc(267) = &HC0851C45: z_Sc(268) = &H7D8B0875: z_Sc(269) = &H6E814
    z_Sc(270) = &HC2610000: z_Sc(271) = &HF6890004: z_Sc(272) = &H850C468B: z_Sc(273) = &H500E74C0: z_Sc(274) = &H4893FF: z_Sc(275) = &H46C70000: z_Sc(276) = &HC&: z_Sc(277) = &H8568B00: z_Sc(278) = &H846C7: z_Sc(279) = &H8B000000: z_Sc(280) = &H890E8B02: z_Sc(281) = &H6C701: z_Sc(282) = &H8B000000: z_Sc(283) = &H4E8B0442: z_Sc(284) = &HC7018904: z_Sc(285) = &H446&: z_Sc(286) = &H8B520000: z_Sc(287) = &H7E30C4A: z_Sc(288) = &H4C93FF51: z_Sc(289) = &HFF000000: z_Sc(290) = &H2C93&: z_Sc(291) = &H44F8B00: z_Sc(292) = &H144D8949: z_Sc(293) = &HC3044F89: z_Sc(294) = &H14838B: z_Sc(295) = &H408B0000: z_Sc(296) = &H83C68908: z_Sc(297) = &H558B0CC6: z_Sc(298) = &H3BC93124: z_Sc(299) = &H3B0B7416
    z_Sc(300) = &H5740456: z_Sc(301) = &HEB10C683: z_Sc(302) = &H7E8B41F2: z_Sc(303) = &HF689C308: z_Sc(304) = &H28733981: z_Sc(305) = &H1B755A75: z_Sc(306) = &H8047981: z_Sc(307) = &H7542B019: z_Sc(308) = &H8798112: z_Sc(309) = &HF1D16219: z_Sc(310) = &H79810975: z_Sc(311) = &HFA3E4A0C: z_Sc(312) = &H3102743A: z_Sc(313) = &HF689C3C9: z_Sc(314) = &H8244C8B: z_Sc(315) = &H410B018B: z_Sc(316) = &H75C08504: z_Sc(317) = &H8798112: z_Sc(318) = &HC0&: z_Sc(319) = &H79810975: z_Sc(320) = &HC&: z_Sc(321) = &HE8077446: z_Sc(322) = &HFFFFFFB4: z_Sc(323) = &H548B16E3: z_Sc(324) = &H448B0C24: z_Sc(325) = &H2890424: z_Sc(326) = &H4204508B: z_Sc(327) = &H31045089: z_Sc(328) = &HCC2C0: z_Sc(329) = &H4002B8
    z_Sc(330) = &H90F6EB80: z_Sc(331) = &H424548B: z_Sc(332) = &H4804428B: z_Sc(333) = &H85044289: z_Sc(334) = &H607075C0: z_Sc(335) = &HBDE8E589: z_Sc(336) = &H8BFFFFFA: z_Sc(337) = &H1483&: z_Sc(338) = &H8788B00: z_Sc(339) = &H748B8B: z_Sc(340) = &H5E30000: z_Sc(341) = &H6FE8&: z_Sc(342) = &H8BFE8900: z_Sc(343) = &H3E9830E: z_Sc(344) = &H8302E9C1: z_Sc(345) = &H3E830CC6: z_Sc(346) = &H51077400: z_Sc(347) = &HFFFECFE8: z_Sc(348) = &HE34959FF: z_Sc(349) = &H10C68305: z_Sc(350) = &H8B8BECEB: z_Sc(351) = &H20&: z_Sc(352) = &HFF5107E3: z_Sc(353) = &H4C93&: z_Sc(354) = &H18B3FF00: z_Sc(355) = &HFF000000: z_Sc(356) = &H3093&: z_Sc(357) = &H81DA8900: z_Sc(358) = &HC8C2&: z_Sc(359) = &H326A5200
    z_Sc(360) = &H6A006A: z_Sc(361) = &H2493FF: z_Sc(362) = &HC2610000: z_Sc(363) = &HF6890004
#If ThunkReleaseTesting = True Then
    z_Sc(364) = &H4C20724E: z_Sc(365) = &H20746665: z_Sc(368) = &H74736554: z_Sc(369) = &H676E69: z_Sc(370) = &H4478B57: z_Sc(371) = &HC781DF89: z_Sc(372) = &H5B0&: z_Sc(373) = &H7D013D: z_Sc(374) = &HC7097C00: z_Sc(375) = &H333E0847: z_Sc(376) = &H51EB4B32: z_Sc(377) = &HCCCCCDBE: z_Sc(378) = &H8C783CC: z_Sc(379) = &H720AF883: z_Sc(380) = &H63F88323: z_Sc(381) = &HEB470377: z_Sc(382) = &H3E73D1B: z_Sc(383) = &H5770000: z_Sc(384) = &HEB02C783: z_Sc(385) = &H270F3D0F: z_Sc(386) = &H5770000: z_Sc(387) = &HEB03C783: z_Sc(388) = &H4C78303: z_Sc(389) = &HE6F7C189
    z_Sc(390) = &H8903EAC1: z_Sc(391) = &H92148DD0: z_Sc(392) = &HD012548D: z_Sc(393) = &HF88D129: z_Sc(394) = &H75C0854F: z_Sc(395) = &H81DF89E7: z_Sc(396) = &H5B0C7: z_Sc(397) = &H83FA8900: z_Sc(398) = &HC93110C2: z_Sc(399) = &H51575251: z_Sc(400) = &H7493FF: z_Sc(401) = &HC35F0000
#End If

    ' fill in the thunk's API addresses and global values
    z_Sc(5) = hCanary                   ' object owning this thunk
    z_Sc(6) = hToken                    ' GDI+ token
    z_Sc(7) = m_DPI                     ' system DPI
    If GdipCreateImageAttributes(z_Sc(8)) = 0 Then
        ' create default ImageAttributes handle
        GdipSetImageAttributesWrapMode z_Sc(8), WrapModeTileFlipXY, 0, 0
    End If                              ' APIs used by the thunk
    hMod = GetModuleHandleA("user32.dll")
    z_Sc(9) = GetProcAddress(hMod, "SetTimer")
    z_Sc(10) = GetProcAddress(hMod, "KillTimer")
#If ThunkReleaseTesting = True Then     ' last API position in thunk
    z_Sc(29) = GetProcAddress(hMod, "MessageBoxA")
    ' zero value prevents jumping to thunk code that won't exist
    ' when ThunkReleaseTesting=False, ensures value is zero
#End If
    z_Sc(11) = GetProcAddress(GetModuleHandleA("ole32.dll"), "CoTaskMemFree")
    hMod = GetModuleHandleA("gdiplus.dll")
    z_Sc(12) = GetProcAddress(hMod, "GdiplusShutdown")
    z_Sc(13) = GetProcAddress(hMod, "GdipSetInterpolationMode")
    z_Sc(14) = GetProcAddress(hMod, "GdipCreateBitmapFromHBITMAP")
    z_Sc(15) = GetProcAddress(hMod, "GdipCreateFromHDC")
    z_Sc(16) = GetProcAddress(hMod, "GdipDeleteGraphics")
    z_Sc(17) = GetProcAddress(hMod, "GdipDrawImageRectRectI")
    z_Sc(18) = GetProcAddress(hMod, "GdipDisposeImage")
    z_Sc(19) = GetProcAddress(hMod, "GdipDisposeImageAttributes")
    ' next 3 APIs only used for navigation-enabled GIFs
    z_Sc(20) = GetProcAddress(hMod, "GdipImageSelectActiveFrame")
    z_Sc(21) = GetProcAddress(hMod, "GdipGetImageWidth")
    z_Sc(22) = GetProcAddress(hMod, "GdipGetImageHeight")
    hMod = GetModuleHandleA("kernel32.dll")
    z_Sc(23) = GetProcAddress(hMod, "MulDiv") ' pixels=MulDiv(Himetric,SysDPI,2540)
    z_Sc(24) = GetProcAddress(hMod, "VirtualFree")
    ' GIF Page Dimension GUID (16 bytes) {6AEDBD6D-3FB5-418A-83A6-7F45229DC872}
    ' used by API GdipImageSelectActiveFrame
    z_Sc(25) = &H6AEDBD6D: z_Sc(26) = &H418A3FB5
    z_Sc(27) = &H457FA683: z_Sc(28) = &H72C89D22
    
    ' release overlay and set the canary's 3 VTable entries
    pvSetOverlay uSafeArray(), VarPtrArray(z_Sc), 0, 0
    pUnkQI = ((hThunk Xor &H80000000) + CanaryQI) Xor &H80000000
    pUnkAddRef = ((hThunk Xor &H80000000) + CanaryAddRef) Xor &H80000000
    pUnkRelease = ((hThunk Xor &H80000000) + CanaryRelease) Xor &H80000000
    pvCreateThunk = hThunk
    
End Function

Private Function pvCreateCanary(theCanary As stdole.IUnknown) As Boolean

    ' The canary is a simple 6 element array that is the canary's ObjPtr
    ' element #0: VTable address which points to element #3
    ' element #1: Canary ref count
    ' element #2: client list (hooked pictures)
    ' element #3: thunk address for IUnknown:QueryInterface
    ' element #4: thunk address for IUnknown:AddRef
    ' element #5: thunk address for IUnknown:Release

    Dim hCanary As Long, hClients As Long, hThunk As Long
    Dim aData() As Long, oCanary As stdole.IUnknown
    Const CLIENT_SIZE As Long = 60  ' initialized for 3 pictures
    
    hCanary = CoTaskMemAlloc(24)
    If hCanary = 0 Then Exit Function
    hClients = CoTaskMemAlloc(CLIENT_SIZE)
    If hClients = 0 Then GoTo abortRoutine
    
    ReDim aData(0 To 5)
    aData(0) = ((hCanary Xor &H80000000) + 12) Xor &H80000000
    aData(1) = 1
    aData(2) = hClients
    hThunk = pvCreateThunk(hCanary, aData(3), aData(4), aData(5))
    If hThunk = 0 Then GoTo abortRoutine
    CopyMemory ByVal hCanary, aData(0), 24
    
    ReDim aData(0 To CLIENT_SIZE \ 4 - 1)
    aData(0) = CLIENT_SIZE \ 4: aData(2) = hThunk
    CopyMemory ByVal hClients, aData(0), CLIENT_SIZE
    CopyMemory oCanary, hCanary, 4
    Set theCanary = oCanary: Set oCanary = Nothing
    pvCreateCanary = True
    Exit Function
    
abortRoutine:
    If hCanary <> 0 Then CoTaskMemFree hCanary
    If hClients <> 0 Then CoTaskMemFree hClients

End Function

Private Function pvSetImageAttributes(tPic As StdPicture, hAttrs As Long, bReplace As Boolean) As Long

    ' returns, resets, or replaces attributes for managed pictures

    Dim hAddr As Long, gAttrs As Long, lValue As Long
    
    hAddr = pvIsManaged(tPic)
    If hAddr = 0 Then Exit Function         ' picture not managed
    
    CopyMemory lValue, ByVal ((hAddr Xor &H80000000) + Offset_ClientPropB) Xor &H80000000, 4
    If bReplace = False Then
        pvSetImageAttributes = lValue
    Else
        If hAttrs <> 0 Then
            GdipCloneImageAttributes hAttrs, gAttrs
            If gAttrs = 0 Then Exit Function
        End If
        If lValue <> 0 Then GdipDisposeImageAttributes lValue
        CopyMemory ByVal ((hAddr Xor &H80000000) + Offset_ClientPropB) Xor &H80000000, gAttrs, 4
    End If

End Function

'///// end of thunk-only related private methods
#End If

Private Sub Class_Terminate()
    Set m_Canary = Nothing
    If m_GdipToken <> 0 Then GdiplusShutdown m_GdipToken
End Sub

Private Sub Class_Initialize()
    m_DPI = pvGetSysDPI()
End Sub

